import abc
import functools

import numpy as np
import scipy as sp


class Problem(abc.ABC):
    """
    Base class used to formulate the problem of minimizing a smooth objective
    :math:`f : \\mathbb{R}^{n} \\to \\mathbb{R}`
    over a set of smooth
    nonlinear constraints
    :math:`c : \\mathbb{R}^{n} \\to \\mathbb{R}^{m}`
    bounded by :math:`l, u \\in \\mathbb{R}^{m}`,
    and variable bounds :math:`x_l, x_u \\in \\mathbb{R}^{n}`,
    yielding the problem
     .. math::
        \\begin{align}
            \\min_{x \\in \\mathbb{R}^{n}} \\quad & f(x) \\\\
            \\text{s.t.} \\quad & l \\leq c(x) \\leq u \\\\
                              & l^x \\leq x \\leq u^x
        \\end{align}

    The Lagrangian of this problem is given by
     .. math::
        \\mathcal{L}(x, y) = f(x) + y^{T} c(x),

    where :math:`y \\in \\mathbb{R}^{m}` is a vector of Lagrange multipliers
    """

    def __init__(self, var_lb: np.ndarray, var_ub: np.ndarray, **args) -> None:
        """
        Creates the problem

        Parameters
        ----------
        var_lb, var_ub : np.ndarray
            The variable bounds :math:`l^x, u^x \\in \\mathbb{R}^{n}`
            constraining the problem
        **args : dict, optional
            Additional arguments to pass to the problem:

            * ``cons_lb`` : np.ndarray
                The lower bounds :math:`l \\in \\mathbb{R}^{m}`
                on the constraints
            * ``cons_ub`` : np.ndarray
                The upper bounds :math:`u \\in \\mathbb{R}^{m}`
                on the constraints
            * ``num_cons`` : int
                The number of constraints (if not given by ``cons_lb`` or ``cons_ub``),
                defaults to 0 (no constraints)
        """
        assert var_lb.shape == var_ub.shape
        assert var_lb.ndim == 1

        assert (var_lb <= var_ub).all()
        assert (var_lb < np.inf).all()
        assert (var_ub > -np.inf).all()

        self.var_lb = np.copy(var_lb)
        self.var_ub = np.copy(var_ub)

        num_cons = args.get("num_cons", None)
        cons_lb = args.get("cons_lb", None)
        cons_ub = args.get("cons_ub", None)

        if cons_lb is not None or cons_ub is not None:
            assert num_cons is None

            if cons_lb is not None:
                (num_cons,) = cons_lb.shape
            else:
                (num_cons,) = cons_ub.shape

            if cons_lb is None:
                cons_lb = np.zeros((num_cons,))
            if cons_ub is None:
                cons_ub = np.zeros((num_cons,))

            assert (cons_lb <= cons_ub).all()
            assert (cons_lb < np.inf).all()
            assert (cons_ub > -np.inf).all()

        else:
            if num_cons is None:
                num_cons = 0

            cons_lb = np.zeros((num_cons,))
            cons_ub = np.zeros((num_cons,))

        self.num_cons = num_cons
        self.cons_lb = cons_lb
        self.cons_ub = cons_ub

    @functools.cached_property
    def var_bounded(self):
        """
        Whether of not the variable bounds are trivial (i.e., all infinte)
        """
        return np.isfinite(self.var_lb).any() or np.isfinite(self.var_ub).any()

    @property
    def num_vars(self) -> int:
        """
        The number of variables in the problem
        """
        (num_vars,) = self.var_lb.shape

        return num_vars

    @abc.abstractmethod
    def obj(self, x: np.ndarray) -> float:
        """
        Parameters
        ----------
        x : np.ndarray
            The primal point :math:`x \\in \\mathbb{R}^{n}` at which
            to evaluate the objective function

        Returns
        -------
        float
            The objective function value :math:`f(x)` at the given primal point :math:`x`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def obj_grad(self, x: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        x : np.ndarray
           The primal point :math:`x \\in \\mathbb{R}^{n}` at which to
           evaluate the objective function gradient

        Returns
        -------
        np.ndarray
            The objective function gradient :math:`\\nabla f(x)`
            at the given primal point :math:`x`
        """
        raise NotImplementedError()

    def cons(self, x: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        x : np.ndarray
           The primal point :math:`x \\in \\mathbb{R}^{n}` at which to
           evaluate the constraint function :math:`c`

        Returns
        -------
        np.ndarray
            The constraint value :math:`c(x)` at the given primal point :math:`x`
        """
        raise NotImplementedError()

    def cons_jac(self, x: np.ndarray) -> sp.sparse.spmatrix:
        """
        Parameters
        ----------
        x : np.ndarray
           The primal point :math:`x \\in \\mathbb{R}^{n}` at which to
           evaluate the constraint Jacobian :math:`J_c`

        Returns
        -------
        sp.sparse.spmatrix
            The constraint Jacobian :math:`J_c(x)` at the given primal point :math:`x`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def lag_hess(self, x: np.ndarray, y: np.ndarray) -> sp.sparse.spmatrix:
        """
        Parameters
        ----------
        x, y : np.ndarray
           The primal / dual points :math:`x \\in \\mathbb{R}^{n}` and
           :math:`y \\in \\mathbb{R}^{m}` at which to
           evaluate the Hessian :math:`\\nabla_{xx} \\mathcal{L}(x, y)`

        Returns
        -------
        sp.sparse.spmatrix
            The matrix
            :math:`\\nabla_{xx} \\mathcal{L}(x, y) \\in \\mathbb{R}^{n \\times n}`
            at the given primal / dual points :math:`x` / :math:`y`
        """
        raise NotImplementedError()



import abc
import logging
from typing import Optional

import numpy as np

from pygradflow.display import StateData, inner_display
from pygradflow.eval import EvalError
from pygradflow.implicit_func import ImplicitFunc
from pygradflow.iterate import Iterate
from pygradflow.log import logger
from pygradflow.params import Params, StepControlType
from pygradflow.problem import Problem
from pygradflow.step.solver.step_solver import StepResult
from pygradflow.step.step_solver_error import StepSolverError
from pygradflow.timer import Timer


class StepControlResult:
    def __init__(
        self,
        iterate: Iterate,
        lamb: float,
        active_set,
        rcond: Optional[float],
        accepted: bool,
    ) -> None:
        self.iterate = iterate
        self.lamb = lamb
        self.active_set = active_set
        self.rcond = rcond
        self.accepted = accepted

    @staticmethod
    def from_step_result(
        step_result: StepResult, lamb: float, accepted: bool
    ) -> "StepControlResult":
        return StepControlResult(
            step_result.iterate,
            lamb,
            step_result.active_set,
            step_result.rcond,
            accepted,
        )


class StepController(abc.ABC):
    def __init__(self, problem: Problem, params: Params) -> None:
        self.problem = problem
        self.params = params
        self.lamb = params.lamb_init

    @abc.abstractmethod
    def step(
        self,
        iterate: Iterate,
        rho: float,
        dt: float,
        display: bool,
        timer: Timer,
    ) -> StepControlResult:
        raise NotImplementedError()

    def update_stepsize_after_fail(self, lamb: float) -> float:
        return 2.0 * lamb

    def compute_step(
        self,
        iterate: Iterate,
        rho: float,
        dt: float,
        display: bool,
        timer: Timer,
    ) -> StepControlResult:
        """
        Computes next step using the step method, handling
        linear algebra errors by reducing the step size.
        """

        def fail_result():
            lamb = 1.0 / dt
            lamb = self.update_stepsize_after_fail(lamb)
            return StepControlResult(iterate, lamb, None, None, False)

        try:
            self.display = None
            if display:

                def res_func(next_iterate):
                    func = ImplicitFunc(self.problem, iterate, dt)
                    return np.linalg.norm(func.value_at(next_iterate, rho))

                self.res_func = res_func
                self.display = inner_display(self.problem, self.params)
                logger.debug("     %s", self.display.header)
            step = self.step(iterate, rho, dt, display, timer)

            if step.accepted:
                step.iterate.check_eval()

            return step
        except StepSolverError as e:
            logger.warning("Step solver error during step computation: %s", e)
            return fail_result()
        except EvalError as e:
            logger.warning("Evaluation error during step computation: %s", e)
            return fail_result()

    def display_step(self, iteration, step):
        level = logger.getEffectiveLevel()
        if not self.display or level > logging.DEBUG:
            return

        iterate = step.iterate
        state = StateData()
        state["iter"] = lambda: iteration
        state["residuum"] = lambda: self.res_func(iterate)
        state["dist"] = lambda: step.diff
        state["active_set_size"] = lambda: step.active_set.sum()
        logger.debug("     %s", self.display.row(state))


def step_controller(problem: Problem, params: Params) -> StepController:
    step_control_type = params.step_control_type

    if step_control_type == StepControlType.Exact:
        from pygradflow.step.exact_control import ExactController

        return ExactController(problem, params)
    elif step_control_type == StepControlType.Fixed:
        from pygradflow.step.fixed_control import FixedStepSizeController

        return FixedStepSizeController(problem, params)
    elif step_control_type == StepControlType.ResiduumRatio:
        from pygradflow.step.residuum_ratio_control import ResiduumRatioController

        return ResiduumRatioController(problem, params)
    elif step_control_type == StepControlType.Optimizing:
        from pygradflow.step.opti_control import OptimizingController

        return OptimizingController(problem, params)
    elif step_control_type == StepControlType.BoxReduced:
        from pygradflow.step.box_control import BoxReducedController

        return BoxReducedController(problem, params)
    else:
        assert step_control_type == StepControlType.DistanceRatio
        from pygradflow.step.distance_ratio_control import DistanceRatioController

        return DistanceRatioController(problem, params)



from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.step.newton_control import NewtonController
from pygradflow.step.step_control import StepControlResult


class FixedStepSizeController(NewtonController):
    def __init__(self, problem: Problem, params: Params) -> None:
        super().__init__(problem, params)
        self.lamb = params.lamb_init

    def step(self, iterate, rho, dt, display, timer):
        assert dt > 0.0

        next_steps = self.newton_steps(iterate, rho, dt)

        step = next(next_steps)

        return StepControlResult.from_step_result(step, self.lamb, True)



import dataclasses
import enum
import typing
from dataclasses import dataclass
from enum import Enum, Flag, auto
from typing import Any, Callable, Optional

if typing.TYPE_CHECKING:
    from .scale import Scaling

import numpy as np


class ActiveSetType(Enum):
    Standard = auto()
    SmallestActiveSet = auto()
    LargestActiveSet = auto()


class NewtonType(Enum):
    """
    The Newton method to be used to solve the semi-smooth systems.
    Different methods have trade-offs in terms of computational
    complexity and convergence speed
    """

    Simplified = auto()
    """
    Simplified Newton method, using the same derivative and
    active set throughout computations (cheapest)
    """
    Full = auto()
    """
    Full Newton method, using a new derivative at each step,
    (requiring new derivative evaluations and factorizations)
    """
    ActiveSet = auto()
    """
    Newton method with fixed derivative, but changing active sets.
    Requires refactorizations but no reevaluations of the derivatives
    """

    Globalized = auto()
    """
    Globalizes the Newton method by using an Armijo line search
    """


class StepSolverType(Enum):
    """
    Step solver type to be used throughout computations
    """

    Standard = auto()
    """
    Unscaled and unsymmetric
    """
    Extended = auto()
    """
    Scaled step solver with improved condition
    """
    Symmetric = auto()
    """
    Scaled solver with symmetric matrix
    """
    Asymmetric = auto()
    """
    Scaled solver with asymmetric matrix
    """


class LinearSolverType(Enum):
    """
    Linear solver to be used throughout the computations
    """

    LU = auto()
    """
    LU decomposition
    """
    MINRES = auto()
    """
    Minimal residual method (MINRES), only
    works with symmetric step solver :py:class:`pygradflow.params.StepSolverType.Symmetric`
    """
    GMRES = auto()
    """
    Generalized minimal residual method (GMRES)
    """

    Cholesky = auto()
    """
    Cholesky factorization using scikit-sparse
    """

    MA57 = auto()
    """
    HSL MA57 solver using pyomo
    """

    MUMPS = auto()
    """
    MUMPS solver
    """

    SSIDS = auto()
    """
    SPRAL / SSIDS solver
    """


class StepControlType(Enum):
    Exact = auto()
    Fixed = auto()
    Optimizing = auto()
    BoxReduced = auto()
    ResiduumRatio = auto()
    DistanceRatio = auto()


class PenaltyUpdate(Enum):
    Constant = auto()
    DualNorm = auto()
    DualEquilibration = auto()
    ParetoDecrease = auto()
    ObjectiveFilter = auto()
    LagrangianFilter = auto()


class Precision(Enum):
    """
    Precision to be used in all calculations
    """

    Single = auto()
    """
    Single precision (32 bit)
    """
    Double = auto()
    """
    Double precision (64 bit)
    """


class DerivCheck(Flag):
    """
    How to check for derivatives
    """

    NoCheck = 0
    """
    Disable checks
    """
    CheckFirst = 1 << 0
    """
    Check first derivatives (objective gradient :math:`\\nabla_{x} f(x)` and :math:`J_c(x)`)
    """
    CheckSecond = 1 << 1
    """
    Check Hessian of Lagrangian (:math:`\\nabla_{xx} \\mathcal{L}(x, y)`)
    """
    CheckAll = CheckFirst | CheckSecond


class ScalingType(Enum):
    """
    How to scale the problem
    """

    NoScaling = auto()
    """
    No scaling
    """

    GradJac = auto()
    """
    Scale based on gradient and equilibration of constraint Jacobian
    """

    KKT = auto()
    """
    Compute scales based on the equilibration of the KKT matrix
    """

    Nominal = auto()
    """
    Scaled based on values of the variable and constraint values
    """

    Custom = auto()
    """
    User-provided custom scaling
    """


@dataclass
class Params:
    """
    Parameters used to solve a :py:class:`pygradflow.problem.Problem`
    using a :py:class:`pygradflow.solver.Solver`
    """

    rho: float = 1e-8

    theta_max: float = 0.9
    theta_ref: float = 0.5

    lamb_init: float = 1.0
    # Up to 1e-6 for single precision?
    lamb_min: float = 1e-12
    lamb_max: float = 1e12
    lamb_inc: float = 2.0
    lamb_red: float = 0.5

    K_P: float = 0.2
    K_I: float = 0.005

    opt_tol: float = 1e-6
    lamb_term: float = 1e-8
    active_tol: float = 1e-8

    local_infeas_tol: float = 1e-8

    active_set_type: ActiveSetType = ActiveSetType.Standard
    active_set_method: Optional[Callable[..., float]] = None

    newton_type: NewtonType = NewtonType.Simplified
    newton_tol: float = 1e-8

    step_control_type: StepControlType = StepControlType.DistanceRatio

    step_solver: Optional[Callable[..., Any]] = None
    step_solver_type: StepSolverType = StepSolverType.Symmetric
    linear_solver_type: LinearSolverType = LinearSolverType.LU
    penalty_update: PenaltyUpdate = PenaltyUpdate.DualNorm

    deriv_check: DerivCheck = DerivCheck.NoCheck
    deriv_pert: float = 1e-8
    deriv_tol: float = 1e-4

    precision: Precision = Precision.Double

    scaling_type: ScalingType = ScalingType.NoScaling
    scaling: Optional["Scaling"] = None

    validate_input: bool = True

    iteration_limit: Optional[int] = None
    time_limit: float = float(np.inf)
    display_interval: float = 0.1

    # lower bound on objective function value,
    # used to detect unbounded problems
    obj_lower_limit: float = -1e10

    report_rcond: bool = False
    collect_path: bool = False

    inertia_correction: bool = False

    def __post_init__(self):
        # Convert enum strings to enum values
        for key, attr in self.annotations():
            if isinstance(attr, enum.EnumMeta):
                val = getattr(self, key)
                if isinstance(val, str):
                    setattr(self, key, attr[val])

    @property
    def dtype(self):
        return np.float32 if self.precision == Precision.Single else np.float64

    def write(self, filename):
        import yaml

        class Dumper(yaml.SafeDumper):
            def __init__(self, stream, **args):
                super().__init__(stream, **args)

            def represent_data(self, data):
                if isinstance(data, Enum):
                    return self.represent_data(data.name)
                return super().represent_data(data)

        with open(filename, "w") as f:
            yaml.dump(dataclasses.asdict(self), f, Dumper=Dumper)

    def annotations(self):
        return type(self).__annotations__.items()

    @staticmethod
    def read(filename):
        import yaml

        with open(filename, "r") as f:
            data = yaml.safe_load(f)
            return Params(**data)



from abc import ABC, abstractmethod
from typing import List, Literal

import numpy as np
from termcolor import colored

from pygradflow.log import logger
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.timer import SimpleTimer


class StateData:
    def __init__(self):
        self._entries = dict()

    def __setitem__(self, key, value):
        self._entries[key] = value

    def __getitem__(self, key):
        try:
            entry = self._entries[key]
            if callable(entry):
                return entry()
            return entry
        except Exception:
            return None


class Format:
    @staticmethod
    def bold(s: str) -> str:
        return colored(s, attrs=["bold"])

    @staticmethod
    def _cond_color(cond: bool) -> Literal["red", "green"]:
        return "green" if cond else "red"

    @staticmethod
    def redgreen(s: str, cond: bool, bold: bool) -> str:
        if bold:
            return colored(s, Format._cond_color(cond), attrs=["bold"])

        return colored(s, Format._cond_color(cond))


class BoldFormatter:
    def __init__(self, format: str):
        self.format = format

    def __call__(self, state):
        return Format.bold(self.format.format(state))


class StringFormatter:
    def __init__(self, format: str):
        self.format = format

    def __call__(self, state):
        return self.format.format(state)


class StepFormatter:
    def __call__(self, state):
        accepted = state
        accept_str = "accepted" if accepted else "rejected"
        return Format.redgreen(accept_str, accepted, bold=True)


class RCondFormatter:
    def __call__(self, state):
        if state is None:
            return "-"

        return "{:5.0e}".format(state)


class Column(ABC):
    def __init__(self, name: str, width: int):
        self.name = name
        self.width = width

    @property
    def header(self) -> str:
        return "{:^{}s}".format(self.name, self.width)

    @abstractmethod
    def content(self, state, last_state) -> str:
        raise NotImplementedError()


class AttrColumn(Column):
    def __init__(self, name: str, width: int, format, attr):
        if isinstance(format, str):
            self.format = StringFormatter(format)
        else:
            self.format = format

        super().__init__(name, width)

        self.attr = attr

    def content(self, state, _) -> str:
        value = self.attr(state)

        if value is None:
            return "{0:>{1}}".format("n/a", self.width)

        return self.format(self.attr(state))


class Display:
    def __init__(self, cols, interval=None):
        self.cols = cols
        self.interval = interval

        self.timer = None
        if self.interval is not None:
            assert self.interval >= 0
            self.timer = SimpleTimer()
        self.last_state = None

    def should_display(self):
        if self.timer is None:
            return True

        return self.timer.elapsed() >= self.interval

    @property
    def header(self) -> str:
        return " ".join([col.header for col in self.cols])

    def row(self, state) -> str:
        if self.timer is not None:
            self.timer.reset()

        row = " ".join([col.content(state, self.last_state) for col in self.cols])
        self.last_state = state
        return row


class StateAttr:
    def __init__(self, name: str):
        self.name = name

    def __call__(self, state):
        value = state[self.name]
        return value


class IterateAttr:
    def __init__(self, name: str):
        self.name = name

    def __call__(self, state):
        iterate = state["iterate"]
        return getattr(iterate, self.name)


class ActiveSetColumn(Column):
    def __init__(self):
        super().__init__("Active set", 10)

    def empty(self):
        return "{:^{}s}".format("--", self.width)

    def content(self, state, last_state):
        curr_active_set = state["active_set"]
        last_active_set = None

        if (last_state is not None) and (state["iter"] == (last_state["iter"] + 1)):
            last_active_set = last_state["active_set"]

        if curr_active_set is None:
            return self.empty()

        display_curr = False

        if last_active_set is None:
            display_curr = True
        elif (curr_active_set != last_active_set).any():
            display_curr = True

        if display_curr:
            num_active = curr_active_set.sum()
            return "{:^{}d}".format(num_active, self.width)
        else:
            return "{:^{}s}".format("--", self.width)


def iter_cols(problem):
    is_bounded = problem.var_bounded
    cols = []
    cols.append(AttrColumn("Aug Lag", 16, "{:16.8e}", StateAttr("aug_lag")))
    cols.append(AttrColumn("Objective", 16, "{:16.8e}", IterateAttr("obj")))

    if is_bounded:
        cols.append(
            AttrColumn("Bound inf", 16, "{:16.8e}", IterateAttr("bound_violation"))
        )

    if problem.num_cons > 0:
        cols.append(
            AttrColumn("Cons inf", 16, "{:16.8e}", IterateAttr("cons_violation"))
        )

    cols.append(AttrColumn("Dual inf", 16, "{:16.8e}", IterateAttr("stat_res")))

    return cols


def solver_display(problem: Problem, params: Params):
    cols: List[Column] = []

    cols.append(AttrColumn("Iter", 6, BoldFormatter("{:6d}"), StateAttr("iter")))

    cols += iter_cols(problem)

    cols.append(
        AttrColumn("Primal step", 16, "{:16.8e}", StateAttr("primal_step_norm"))
    )
    cols.append(AttrColumn("Dual step", 16, "{:16.8e}", StateAttr("dual_step_norm")))
    cols.append(AttrColumn("Lambda", 16, "{:16.8e}", StateAttr("lamb")))

    if problem.var_bounded:
        cols.append(ActiveSetColumn())

    if params.report_rcond:
        cols.append(AttrColumn("Rcond", 5, RCondFormatter(), StateAttr("rcond")))

    cols.append(AttrColumn("Obj nonlin", 16, "{:16.8e}", StateAttr("obj_nonlin")))

    if problem.num_cons > 0:
        cols.append(AttrColumn("Cons nonlin", 16, "{:16.8e}", StateAttr("cons_nonlin")))

    cols.append(AttrColumn("Type", 8, StepFormatter(), StateAttr("step_accept")))

    return Display(cols, interval=params.display_interval)


class FilterColumn(Column):
    def __init__(self):
        super().__init__("Filter", 10)

    def empty(self):
        return "{:^{}s}".format("--", self.width)

    def content(self, state, last_state):
        curr_filter = state["filter"]
        last_filter = None

        if (last_state is not None) and (state["iter"] == (last_state["iter"] + 1)):
            last_filter = last_state["filter"]

        if curr_filter is None:
            return self.empty()

        display_curr = False

        if last_filter is None:
            display_curr = True
        elif (curr_filter != last_filter).any():
            display_curr = True

        if display_curr:
            num_active = curr_filter.sum()
            return "{:^{}d}".format(num_active, self.width)
        else:
            return "{:^{}s}".format("--", self.width)


class StepTypeColumn(Column):
    def __init__(self):
        super().__init__("Step Type", 10)
        self.format = BoldFormatter("{:^10s}")

    def content(self, state, last_state):
        step_type = state["step_type"]
        return self.format(step_type.name())


def integrator_display(problem: Problem, params: Params):
    cols: List[Column] = []

    cols.append(AttrColumn("Iter", 6, BoldFormatter("{:6d}"), StateAttr("iter")))
    cols += iter_cols(problem)

    cols.append(FilterColumn())

    cols.append(AttrColumn("Func evals", 10, "{:10d}", StateAttr("num_func_evals")))
    cols.append(AttrColumn("Jac evals", 10, "{:10d}", StateAttr("num_jac_evals")))
    cols.append(AttrColumn("Steps", 10, "{:10d}", StateAttr("num_steps")))
    cols.append(AttrColumn("dt", 12, "{:6e}", StateAttr("dt")))

    cols.append(StepTypeColumn())

    return Display(cols)


def inner_display(problem: Problem, params: Params):
    cols: List[Column] = []

    cols.append(AttrColumn("Iter", 6, BoldFormatter("{:5d}"), StateAttr("iter")))
    cols.append(AttrColumn("Residuum", 16, "{:16.8e}", StateAttr("residuum")))
    cols.append(AttrColumn("Dist", 16, "{:16.8e}", StateAttr("dist")))
    cols.append(AttrColumn("Active set", 10, "{:10d}", StateAttr("active_set_size")))

    return Display(cols)


def print_problem_stats(problem, iterate):
    num_vars = problem.num_vars
    num_cons = problem.num_cons

    logger.info("Solving problem with %s variables, %s constraints", num_vars, num_cons)

    cons_jac = iterate.cons_jac
    lag_hess = iterate.lag_hess(iterate.y)

    var_lb = problem.var_lb
    var_ub = problem.var_ub

    inf_lb = var_lb == -np.inf
    inf_ub = var_ub == np.inf

    unbounded = np.logical_and(inf_lb, inf_ub)
    ranged = np.logical_and(np.logical_not(inf_lb), np.logical_not(inf_ub))
    bounded_below = np.logical_and(np.logical_not(inf_lb), inf_ub)
    bounded_above = np.logical_and(inf_lb, np.logical_not(inf_ub))
    bounded = np.logical_or(bounded_below, bounded_above)
    fixed = var_lb[ranged] == var_ub[ranged]

    num_fixed = np.sum(fixed)
    num_unbounded = np.sum(unbounded)
    num_ranged = np.sum(ranged)
    num_bounded = np.sum(bounded)

    has_cons = num_cons > 0

    logger.info("           Hessian nnz: %s", lag_hess.nnz)
    if has_cons:
        logger.info("          Jacobian nnz: %s", cons_jac.nnz)

    logger.info("       Fixed variables: %s", num_fixed)
    logger.info("        Free variables: %s", num_unbounded)
    logger.info("     Bounded variables: %s", num_bounded)
    logger.info("      Ranged variables: %s", num_ranged)

    if not has_cons:
        return

    cons_lb = problem.cons_lb
    cons_ub = problem.cons_ub

    equation = cons_lb == cons_ub
    ranged = np.logical_and(np.isfinite(cons_lb), np.isfinite(cons_ub))
    ranged = np.logical_and(ranged, np.logical_not(equation))

    num_equations = equation.sum()
    num_ranged = ranged.sum()
    num_inequalities = num_cons - num_equations - num_ranged

    logger.info("  Equality constraints: %s", num_equations)
    logger.info("Inequality constraints: %s", num_inequalities)
    logger.info("    Ranged constraints: %s", num_ranged)



import datetime
import enum
import itertools
import logging
import os
from abc import ABC, abstractmethod
from concurrent.futures import FIRST_COMPLETED, ProcessPoolExecutor, wait
from multiprocessing import Process, Queue, cpu_count

import numpy as np

from pygradflow.log import logger
from pygradflow.params import Params
from pygradflow.solver import SolverStatus

run_logger = logging.getLogger(__name__)

formatter = logging.Formatter("%(asctime)s:%(name)s:%(levelname)s:%(message)s")


def solve_instance(instance, params, log_filename, verbose):
    logger.handlers.clear()

    if log_filename is not None:
        handler = logging.FileHandler(log_filename)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if verbose:
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logger.setLevel(logging.INFO)

    logging.captureWarnings(True)
    warn_logger = logging.getLogger("py.warnings")
    warn_logger.addHandler(handler)
    warn_logger.setLevel(logging.WARN)

    try:
        np.seterr(divide="raise", over="raise", invalid="raise")
        result = instance.solve(params)
        return (instance, result)
    except Exception as exc:
        logger.error("Error solving %s", instance.name, exc_info=exc)
        return (instance, "error")


def solve_instance_queue(queue, instance, params, log_filename, verbose):
    instance_result = solve_instance(instance, params, log_filename, verbose)
    queue.put(instance_result)


def solve_instance_time_limit(instance, params, log_filename, verbose):
    if params.time_limit == np.inf:
        return solve_instance(instance, params, log_filename, verbose)

    queue = Queue()
    process = Process(
        target=solve_instance_queue,
        args=(queue, instance, params, log_filename, verbose),
    )

    process.start()
    process.join(timeout=(params.time_limit + 10.0))

    if process.is_alive():
        logger.error("Reached timeout, terminating process for %s", instance.name)
        process.terminate()
        logger.error("Terminated process for %s", instance.name)
        return (instance, "timeout")
    else:
        return queue.get()


class Runner(ABC):
    def __init__(self, name):
        self.name = name

    @abstractmethod
    def get_instances(self, args):
        raise NotImplementedError()

    def create_params(self, args):
        params = Params()

        for key, attr in params.annotations():
            value = getattr(args, key)
            if isinstance(attr, enum.EnumMeta):
                value = attr[value]
            setattr(params, key, value)

        return params

    def log_filename(self, args, instance):
        if args.no_output:
            return None
        return self.output_filename(args, f"{instance.name}.log")

    def solve_instances_sequential(self, instances, args, params):
        verbose = args.verbose
        for instance in instances:
            log_filename = self.log_filename(args, instance)
            yield solve_instance_time_limit(instance, params, log_filename, verbose)

    def solve_instances_parallel(self, instances, args, params):
        verbose = args.verbose

        if args.parallel is True:
            num_procs = cpu_count()
        else:
            num_procs = args.parallel

        run_logger.info("Solving in parallel with up to %d processes", num_procs)

        all_params = itertools.repeat(params)
        all_log_filenames = [
            self.log_filename(args, instance) for instance in instances
        ]
        all_verbose = itertools.repeat(verbose)

        solve_args = zip(instances, all_params, all_log_filenames, all_verbose)

        with ProcessPoolExecutor(num_procs, max_tasks_per_child=1) as pool:

            futures = []

            for solve_arg in solve_args:
                future = pool.submit(solve_instance_time_limit, *solve_arg)
                future.instance = solve_arg[0]
                futures.append(future)

            run_logger.info("Started %d instances", len(futures))

            while True:
                done, futures = wait(futures, return_when=FIRST_COMPLETED)

                for item in done:
                    yield item.result()

                if len(futures) == 0:
                    run_logger.info("Solved all instances, terminating")
                    break

                remaining_future = next(iter(futures))

                run_logger.info(
                    "Finished %d instances, waiting for %d (including %s)",
                    len(done),
                    len(futures),
                    remaining_future.instance.name,
                )

    def solve_instances(self, instances, args):
        # yields sequence of tuples of (instance, result) for each instance
        run_logger.info("Solving %d instances", len(instances))

        params = self.create_params(args)

        if args.parallel is not None:
            yield from self.solve_instances_parallel(instances, args, params)
        else:
            yield from self.solve_instances_sequential(instances, args, params)

    def filter_instances(self, args):
        instances = []

        max_size = args.max_size
        name = args.name

        for instance in self.get_instances(args):
            if max_size is not None and instance.size > max_size:
                continue

            if args.unconstrained and instance.num_cons > 0:
                continue

            if name is not None and name != instance.name:
                continue

            instances.append(instance)

        return instances

    def parser(self):
        import argparse

        parser = argparse.ArgumentParser()

        group = parser.add_argument_group(title="parameters")

        default_params = Params()

        for key, attr in default_params.annotations():
            name = f"--{key}"
            if isinstance(attr, enum.EnumMeta):
                default_value = getattr(default_params, key).name
                group.add_argument(
                    name, default=default_value, type=str, help="Default: %(default)s"
                )
            else:
                default_value = getattr(default_params, key)

                group.add_argument(
                    name,
                    default=default_value,
                    type=attr,
                    help="Default: %(default)s",
                )

        group = parser.add_argument_group(title="runner")

        parser.add_argument("--output", type=str)
        parser.add_argument("--no_output", action="store_true")
        parser.add_argument("--max_size", type=int)
        parser.add_argument("--name", type=str)
        parser.add_argument("--unconstrained", action="store_true")
        parser.add_argument("--verbose", action="store_true")
        parser.add_argument("--parallel", nargs="?", type=int, const=True)

        return parser

    def output_filename(self, args, filename):
        return os.path.join(args.output, filename)

    def main(self):
        run_logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        run_logger.addHandler(handler)

        args = self.parser().parse_args()

        with_output = not (args.no_output)

        if with_output:
            if args.output is None:
                now = datetime.datetime.now().isoformat("T", "seconds")
                args.output = f"output_{self.name}_{now}"

            os.makedirs(args.output, exist_ok=True)

        instances = self.filter_instances(args)

        self.solve(instances, args)

    def create_csv_row(self, args, instance, result):
        info = {
            "instance": instance.name,
            "num_vars": instance.num_vars,
            "num_cons": instance.num_cons,
            "size": instance.size,
        }

        default_props = {
            "iterations": 0,
            "num_accepted_steps": 0,
            "final_scaled_obj": 0.0,
            "final_stat_res": 0.0,
            "final_cons_violation": 0.0,
            "dist_factor": 0.0,
        }

        if result == "timeout":
            return {
                **info,
                "status": "timeout",
                "total_time": args.time_limit,
                **default_props,
            }

        elif result == "error":
            return {**info, "status": "error", "total_time": 0.0, **default_props}
        else:
            return {
                **info,
                "status": SolverStatus.short_name(result.status),
                "total_time": result.total_time,
                "iterations": result.iterations,
                "num_accepted_steps": result.num_accepted_steps,
                "final_scaled_obj": result.final_scaled_obj,
                "final_stat_res": result.final_stat_res,
                "final_cons_violation": result.final_cons_violation,
                "dist_factor": result.dist_factor,
            }

    def solve(self, instances, args):
        import csv

        with_output = not (args.no_output)

        params = self.create_params(args)
        filename = None

        if with_output:
            params.write(self.output_filename(args, "params.yml"))

            filename = self.output_filename(args, "output.csv")

            run_logger.info("Writing results to '%s'", filename)

        fieldnames = [
            "instance",
            "num_vars",
            "num_cons",
            "size",
            "status",
            "total_time",
            "iterations",
            "num_accepted_steps",
            "final_scaled_obj",
            "final_stat_res",
            "final_cons_violation",
            "dist_factor",
        ]

        if with_output:
            with open(filename, "w") as output_file:
                writer = csv.DictWriter(output_file, fieldnames=fieldnames)
                writer.writeheader()

                for instance, result in self.solve_instances(instances, args):
                    run_logger.info("Finished instance %s", instance.name)
                    writer.writerow(self.create_csv_row(args, instance, result))
                    output_file.flush()

        else:
            for instance, result in self.solve_instances(instances, args):
                run_logger.info("Finished instance %s", instance.name)



from functools import cached_property

import numpy as np
import pycutest

from pygradflow.problem import Problem

from .instance import Instance
from .runner import Runner


def cutest_is_ne_prob(name):
    return name.endswith("NE")


def cutest_map_inf(x):
    cutest_inf = 1e20
    y = np.copy(x)
    y[y == cutest_inf] = np.inf
    y[y == -cutest_inf] = -np.inf
    return y


class UnconstrainedCUTEstProblem(Problem):
    def __init__(self, instance):
        self.instance = instance
        var_lb = cutest_map_inf(instance.bl)
        var_ub = cutest_map_inf(instance.bu)
        assert instance.m == 0
        super().__init__(var_lb, var_ub, num_cons=0)

    def obj(self, x):
        return self.instance.obj(x)

    def obj_grad(self, x):
        obj, grad = self.instance.obj(x, gradient=True)
        return grad

    def cons(self, x):
        raise NotImplementedError("Unconstrained problem")

    def cons_jac(self, x):
        raise NotImplementedError("Unconstrained problem")

    def lag_hess(self, x, y):
        return self.instance.sphess(x, v=None)

    def x0(self):
        return self.instance.x0


class ConstrainedCUTEstProblem(Problem):
    def __init__(self, instance):
        self.instance = instance
        var_lb = cutest_map_inf(instance.bl)
        var_ub = cutest_map_inf(instance.bu)
        cons_lb = cutest_map_inf(instance.cl)
        cons_ub = cutest_map_inf(instance.cu)

        super().__init__(var_lb, var_ub, cons_lb=cons_lb, cons_ub=cons_ub)

    def obj(self, x):
        return self.instance.obj(x)

    def obj_grad(self, x):
        obj, grad = self.instance.obj(x, gradient=True)
        return grad

    def cons(self, x):
        return self.instance.cons(x)

    def cons_jac(self, x):
        cons, jac = self.instance.scons(x, gradient=True)
        return jac

    def lag_hess(self, x, y):
        return self.instance.sphess(x, v=y)

    @property
    def x0(self):
        return self.instance.x0()


# Nonlinear equations: Goal is to minimize the violation
# 1/2 ||c(x)||^2 subject to problem bounds. Constraint
# functions are used to access the residuals and their derivatives
class NECUTEstProblem(Problem):
    def __init__(self, instance):
        self.instance = instance
        var_lb = cutest_map_inf(instance.bl)
        var_ub = cutest_map_inf(instance.bu)

        super().__init__(var_lb, var_ub, num_cons=0)

    @cached_property
    def problem(self):
        return pycutest.import_problem(self.name, drop_fixed_variables=True)

    def obj(self, x):
        residuals = self.instance.cons(x)
        return 0.5 * np.dot(residuals, residuals)

    def obj_grad(self, x):
        residuals, jac = self.instance.scons(x, gradient=True)
        return jac.T.dot(residuals)

    def lag_hess(self, x, y):
        _, jac = self.instance.scons(x, gradient=True)
        return jac.T.dot(jac)

    @property
    def x0(self):
        return self.instance.x0

    @property
    def y0(self):
        return np.zeros((self.num_cons,))


class CUTestInstance(Instance):
    def __init__(self, instance):
        self.instance = instance

        props = pycutest.problem_properties(self.instance)

        num_vars = props["n"]
        num_cons = props["m"]

        if num_cons is None:
            num_cons = 0

        super().__init__(instance, num_vars, num_cons)

    def cutest_problem(self):
        return pycutest.import_problem(self.name, drop_fixed_variables=True)

    def x0(self):
        cutest_problem = self.cutest_problem()
        return cutest_problem.x0

    def problem(self):
        cutest_problem = self.cutest_problem()

        is_ne = cutest_is_ne_prob(self.name)

        if is_ne:
            return NECUTEstProblem(cutest_problem)
        elif cutest_problem.m == 0:
            return UnconstrainedCUTEstProblem(cutest_problem)
        else:
            return ConstrainedCUTEstProblem(cutest_problem)


class CUTestRunner(Runner):
    def __init__(self):
        super().__init__(name="cutest")

    def parser(self):
        parser = super().parser()

        parser.add_argument("--ignore_ne_probs", action="store_true")

        return parser

    def get_instances(self, args):
        instances = pycutest.find_problems()
        filtered_instances = []

        for instance in instances:
            props = pycutest.problem_properties(instance)

            if args.ignore_ne_probs and cutest_is_ne_prob(instance):
                continue

            if props["m"] == "variable":
                continue

            if props["n"] == "variable":
                continue

            filtered_instances.append(CUTestInstance(instance))

        return filtered_instances


if __name__ == "__main__":
    CUTestRunner().main()



import pyspral.ssids as ssids

from .linear_solver import LinearSolver, LinearSolverError


class SSIDSSolver(LinearSolver):
    def __init__(self, mat, symmetric=False, report_rcond=False):
        super().__init__(mat, symmetric=symmetric)

        try:
            self.symbolic_factor = ssids.analyze(mat, check=True)
            self.numeric_factor = self.symbolic_factor.factor(posdef=False)
        except ssids.SSIDSError as e:
            raise LinearSolverError() from e

    def solve(self, rhs, trans=False, initial_sol=None):
        try:
            return self.numeric_factor.solve(rhs, inplace=False)
        except ssids.SSIDSError as e:
            raise LinearSolverError() from e

    def num_neg_eigvals(self):
        return self.numeric_factor.inform.num_neg



import time


class SimpleTimer:
    def __init__(self):
        self.start = time.time()

    def elapsed(self):
        return time.time() - self.start

    def reset(self):
        self.start = time.time()


class Timer(SimpleTimer):
    def __init__(self, time_limit: float):
        super().__init__()
        self.time_limit = time_limit

    def remaining(self):
        return self.time_limit - self.elapsed()

    def reached_time_limit(self):
        return self.remaining() <= 0.0



from typing import Iterator, Optional

import numpy as np

from pygradflow.iterate import Iterate
from pygradflow.newton import newton_method
from pygradflow.params import ActiveSetType, Params
from pygradflow.problem import Problem
from pygradflow.step.solver.step_solver import StepResult
from pygradflow.step.step_control import StepController


class NewtonController(StepController):
    """
    Step controller working by solving the implicit Euler
    equations using a semi-smooth Newton method
    """

    def __init__(self, problem: Problem, params: Params) -> None:
        super().__init__(problem, params)

    def newton_steps(
        self,
        orig_iterate: Iterate,
        rho: float,
        dt: float,
    ) -> Iterator[StepResult]:
        problem = self.problem
        params = self.params

        tau = self.compute_tau(orig_iterate, rho)
        self.method = newton_method(problem, params, orig_iterate, dt, rho, tau)
        curr_iterate = orig_iterate

        while True:
            next_step = self.method.step(curr_iterate)
            yield next_step
            curr_iterate = next_step.iterate

    def tau_vals(self, initial_iterate: Iterate, rho: float):
        problem = self.problem

        x = initial_iterate.x
        g = initial_iterate.aug_lag_deriv_x(rho)
        xl = problem.var_lb
        xu = problem.var_ub

        zero_g = np.isclose(g, 0.0)
        nonzero_g = np.logical_not(zero_g)
        pos_g = np.logical_and(g > 0.0, nonzero_g)
        neg_g = np.logical_and(g < 0.0, nonzero_g)

        tau_vals = np.full_like(x, fill_value=-1)

        tau_vals[pos_g] = (x[pos_g] - xl[pos_g]) / g[pos_g]
        tau_vals[neg_g] = (xu[neg_g] - x[neg_g]) / -g[neg_g]

        return tau_vals

    def compute_tau(self, initial_iterate: Iterate, rho: float) -> Optional[float]:
        params = self.params

        active_set_method = params.active_set_method

        if active_set_method is not None:
            return active_set_method(initial_iterate, self.lamb, rho)

        active_set_type = params.active_set_type

        if active_set_type == ActiveSetType.Standard:
            return None

        tau_vals = self.tau_vals(initial_iterate, rho)

        if active_set_type == ActiveSetType.SmallestActiveSet:
            if (tau_vals <= 0).all():
                return 1.0

            min_tau = np.min(tau_vals[tau_vals > 0])
            assert min_tau >= 0
            return 0.5 * min_tau

        return max(np.max(tau_vals), 1.0)



import numpy as np
import scipy as sp

from pygradflow.problem import Problem
from pygradflow.util import sparse_zero


class ConstrainedProblem(Problem):
    def __init__(self, problem):
        cons_lb = problem.cons_lb
        cons_ub = problem.cons_ub
        self.problem = problem
        self.create_slacks(cons_lb, cons_ub)

        var_lb = problem.var_lb
        var_ub = problem.var_ub

        num_slacks = len(self.slack_positions)

        slack_lb = cons_lb[self.slack_positions]
        slack_ub = cons_ub[self.slack_positions]

        if num_slacks > 0:
            var_lb = np.concatenate([var_lb, slack_lb])

            var_ub = np.concatenate([var_ub, slack_ub])

        num_cons = problem.num_cons
        super().__init__(var_lb, var_ub, num_cons=num_cons)

    def create_slacks(self, cons_lb, cons_ub):
        (num_cons,) = cons_lb.shape
        assert cons_lb.shape == cons_ub.shape

        has_offsets = False
        cons_offsets = np.zeros((num_cons,))

        slack_positions = []

        for i, (lb, ub) in enumerate(zip(cons_lb, cons_ub)):
            if lb == ub:
                if lb != 0.0:
                    has_offsets = True
                    cons_offsets[i] = -lb
                continue

            slack_positions.append(i)

        self.slack_positions = np.array(slack_positions, dtype=int)

        self.cons_offsets = None

        if has_offsets:
            self.cons_offsets = cons_offsets

    def orig_vals(self, x):
        orig_num_vars = self.problem.num_vars
        return x[:orig_num_vars]

    def slack_vals(self, x):
        orig_num_vars = self.problem.num_vars
        return x[orig_num_vars:]

    def obj(self, x):
        return self.problem.obj(self.orig_vals(x))

    def obj_grad(self, x):
        orig_grad = self.problem.obj_grad(self.orig_vals(x))

        num_slacks = len(self.slack_positions)

        if num_slacks == 0:
            return orig_grad

        return np.concatenate([orig_grad, np.zeros((num_slacks,))])

    def cons(self, x):
        orig_cons = self.problem.cons(self.orig_vals(x))

        num_slacks = len(self.slack_positions)

        if self.cons_offsets is not None:
            orig_cons += self.cons_offsets

        if num_slacks == 0:
            return orig_cons

        slack_vals = self.slack_vals(x)

        for pos, val in zip(self.slack_positions, slack_vals):
            orig_cons[pos] -= val

        return orig_cons

    def cons_jac(self, x):
        orig_cons_jac = self.problem.cons_jac(self.orig_vals(x))

        num_slacks = len(self.slack_positions)

        if num_slacks == 0:
            return orig_cons_jac

        data = np.full((num_slacks,), fill_value=-1.0, dtype=float)
        cols = np.arange(num_slacks)
        rows = self.slack_positions
        num_cons = self.num_cons

        additional_cons_jac = sp.sparse.coo_matrix(
            (data, (rows, cols)), shape=(num_cons, num_slacks)
        )

        return sp.sparse.bmat([[orig_cons_jac, additional_cons_jac]])

    def lag_hess(self, x, y):
        orig_hess = self.problem.lag_hess(self.orig_vals(x), y)

        (num_vars, _) = orig_hess.shape

        num_slacks = len(self.slack_positions)

        if num_slacks == 0:
            return orig_hess

        lower = sparse_zero(shape=(num_slacks, num_vars))
        diag = sparse_zero(shape=(num_slacks, num_slacks))

        return sp.sparse.bmat([[orig_hess, lower.T], [lower, diag]])

    def transform_sol(self, orig_x, orig_y):
        assert orig_x.shape == (self.problem.num_vars,)
        assert orig_y.shape == (self.problem.num_cons,)
        problem = self.problem

        num_slacks = len(self.slack_positions)

        if num_slacks == 0:
            return (orig_x, orig_y)

        slack_vals = np.zeros((num_slacks,))

        orig_cons_vals = self.problem.cons(orig_x)

        # Ensure that initial solution satisfies variable bounds
        # of transformed problem even if constraints at
        # initial point are violated
        for i, pos in enumerate(self.slack_positions):
            cons_val = orig_cons_vals[pos]
            lb_val = problem.cons_lb[pos]
            ub_val = problem.cons_ub[pos]

            slack_val = np.clip(cons_val, lb_val, ub_val)
            slack_vals[i] = slack_val

        x = np.concatenate([orig_x, slack_vals])
        y = orig_y

        return (x, y)

    def restore_sol(self, x, y, d):
        assert x.shape == (self.num_vars,)
        assert y.shape == (self.num_cons,)
        assert d.shape == (self.num_vars,)

        num_slacks = len(self.slack_positions)

        if num_slacks == 0:
            return (x, y, d)

        orig_x = self.orig_vals(x)
        orig_y = y
        orig_d = self.orig_vals(d)

        return (orig_x, orig_y, orig_d)



import numpy as np

from pygradflow.problem import Problem
from pygradflow.status import SolverStatus


class SolverResult:
    """
    The result of a solution of a :py:class:`pygradflow.problem.Problem`
    instance with a :py:class:`pygradflow.solver.Solver`
    """

    def __init__(
        self,
        problem: Problem,
        x: np.ndarray,
        y: np.ndarray,
        d: np.ndarray,
        status: SolverStatus,
        iterations: int,
        num_accepted_steps: int,
        total_time: float,
        dist_factor: float,
        **attrs
    ):
        self.problem = problem
        self._attrs = attrs

        self._x = x
        self._y = y
        self._d = d
        self._status = status
        self.iterations = iterations
        self.num_accepted_steps = num_accepted_steps
        self.total_time = total_time
        self.dist_factor = dist_factor

    def _set_path(self, path, model_times):
        self._attrs["path"] = path
        self._attrs["model_times"] = model_times

        num_vars = self.problem.num_vars
        num_cons = self.problem.num_cons

        assert model_times.ndim == 1
        assert path.shape == (num_vars + num_cons, len(model_times))

        self._attrs["primal_path"] = lambda: path[:num_vars]
        self._attrs["dual_path"] = lambda: path[num_vars:]

        def speed():
            return np.linalg.norm(np.diff(self.path, axis=1), axis=0) / np.diff(
                model_times
            )

        def primal_speed():
            return np.linalg.norm(np.diff(self.primal_path, axis=1), axis=0) / np.diff(
                model_times
            )

        def dual_speed():
            return np.linalg.norm(np.diff(self.dual_path, axis=1), axis=0) / np.diff(
                model_times
            )

        self._attrs["model_speed"] = speed
        self._attrs["primal_model_speed"] = primal_speed
        self._attrs["dual_model_speed"] = dual_speed

    @property
    def status(self) -> SolverStatus:
        """
        The status of the solve as a :py:class:`pygradflow.solver.SolverStatus`
        """
        return self._status

    def __getattr__(self, name):
        attrs = super().__getattribute__("_attrs")
        val = attrs.get(name, None)

        if val is None:
            return val

        if callable(val):
            return val()

        return val

    def __setitem__(self, name, value):
        self._attrs[name] = value

    def __getitem__(self, name, value):
        return self._attrs[name]

    @property
    def x(self) -> np.ndarray:
        """
        The primal solution :math:`x \\in \\mathbb{R}^{n}`
        """
        return self._x

    @property
    def y(self) -> np.ndarray:
        """
        The dual solution :math:`y \\in \\mathbb{R}^{m}`
        """
        return self._y

    @property
    def d(self) -> np.ndarray:
        """
        The dual solution :math:`d \\in \\mathbb{R}^{n}`
        with respect to the variable bounds
        """
        return self._d

    def __repr__(self) -> str:
        return "SolverResult(status={0})".format(self.status)

    @property
    def success(self):
        return SolverStatus.success(self.status)



import numpy as np
import scipy as sp

from pygradflow.iterate import Iterate
from pygradflow.linear_solver import LinearSolverError
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.step.step_solver_error import StepSolverError

from .scaled_step_solver import ScaledStepSolver


class ExtendedStepSolver(ScaledStepSolver):
    """
    Extended step solver. Computes Newton step based on the
    scaled implicit function and a modified formulation in
    order to improve condition and sparsity.
    """

    def __init__(
        self,
        problem: Problem,
        params: Params,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
    ) -> None:
        super().__init__(problem, params, orig_iterate, dt, rho)

        assert dt > 0.0
        assert rho > 0.0

    def extract_rows(
        self, mat: sp.sparse.spmatrix, row_filter: np.ndarray
    ) -> sp.sparse.spmatrix:
        mat = mat.tocsc()
        return mat[row_filter, :]

    def _compute_deriv(self) -> None:
        active_indices = np.where(self.active_set)[0]
        inactive_indices = np.where(np.logical_not(self.active_set))[0]

        assert self.dt > 0.0

        lamb = 1.0 / self.dt
        rho = self.rho

        active_set_size = active_indices.size

        n = self.n
        m = self.m

        active_rows = np.arange(active_set_size)
        active_data = np.ones((active_set_size,), dtype=self.params.dtype)

        trans_active_mat = sp.sparse.coo_matrix(
            (active_data, (active_rows, active_indices)), shape=(active_set_size, n)
        )

        lower_mat = sp.sparse.diags(
            [-lamb / (1.0 + lamb * rho)], shape=(m, m), dtype=self.params.dtype
        )

        jac = self.jac
        hess = self.hess + sp.sparse.diags(
            [lamb], shape=(n, n), dtype=self.params.dtype
        )

        filtered_hess = self.extract_rows(hess, inactive_indices)
        trans_filtered_jac = self.extract_rows(jac.T, inactive_indices)

        blocks = [
            [trans_active_mat, None],
            [filtered_hess, trans_filtered_jac],
            [jac, lower_mat],
        ]

        self._deriv = sp.sparse.bmat(blocks, format="csc")

        assert self.deriv.shape == (n + m, n + m)
        assert self.deriv.dtype == self.params.dtype

    def solve_scaled(self, b0, b1, b2t):
        params = self.params
        if self._deriv is None:
            self._compute_deriv()

        n = self.n
        m = self.m

        rhs = np.concatenate((b0, b1, b2t))

        assert rhs.shape == (n + m,)

        try:
            if self.solver is None:
                self.solver = self.linear_solver(self.deriv)
            sol = self.solver.solve(rhs)
        except LinearSolverError as e:
            raise StepSolverError from e

        dx = sol[:n]
        dy = sol[n:]

        rcond = None
        if params.report_rcond:
            try:
                rcond = self.estimate_rcond(self.deriv, self.solver)
            except LinearSolverError:
                pass

        return (dx, dy, rcond)



import math

import cyipopt
import numpy as np

from pygradflow.eval import EvalError, astype
from pygradflow.iterate import Iterate
from pygradflow.step.step_control import (
    StepController,
    StepControlResult,
    StepSolverError,
)


def map_cython_exception(func):
    def wrapped_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (ArithmeticError, EvalError) as e:
            raise cyipopt.CyIpoptEvaluationError() from e

    return wrapped_func


class ImplicitProblem(cyipopt.Problem):
    """
    Solves the primal-dual proximally regularized problem
     .. math::
        \\begin{align}
            \\min_{x \\in \\mathbb{R}^{n}, w \\in \\mathbb{R}^{n}} \\quad
            & f(x) + \\frac{\\rho}{2} \\|c(x)\\|^{2} + \\frac{\\lambda}{2} (\\| x - \\hat{x} \\|^{2} + \\| w - \\hat{y} \\|^{2}) \\\\
            \\text{s.t.} \\quad & 0 = c(x) + \\lambda w \\\\
                              & l^x \\leq x \\leq u^x
        \\end{align}

    whose optimum coincides with the implicit Euler step of the augmented Lagrangian
    of the original problem. The problem is solved using Ipopt with a limited-memory
    BFGS Hessian approximation.
    """

    def __init__(self, iterate, lamb, rho, rescaled=True):
        # Resacled:
        # Replace w by v = w / sqrt(lambda) in the augmented Lagrangian
        # to avoid numerical issues
        assert lamb >= 0
        assert rho >= 0

        self.iterate = iterate
        self.eval = iterate.eval
        self.lamb = lamb
        self.rho = rho

        self.rescaled = rescaled

        problem = iterate.problem
        prob_num_vars = problem.num_vars
        prob_num_cons = problem.num_cons

        num_vars = prob_num_vars + prob_num_cons
        num_cons = prob_num_cons

        self.var_lb = np.concatenate([problem.var_lb, np.full(prob_num_cons, -np.inf)])
        self.var_ub = np.concatenate([problem.var_ub, np.full(prob_num_cons, np.inf)])

        self.cons_lb = np.zeros(prob_num_cons)
        self.cons_ub = np.zeros(prob_num_cons)

        self.problem = problem
        self.prob_num_vars = prob_num_vars
        self.prob_num_cons = prob_num_cons

        self.num_vars = num_vars
        self.num_cons = num_cons

        super(ImplicitProblem, self).__init__(
            n=num_vars,
            m=num_cons,
            lb=self.var_lb,
            ub=self.var_ub,
            cl=self.cons_lb,
            cu=self.cons_ub,
        )

    @map_cython_exception
    def objective(self, z):
        assert z.shape == (self.num_vars,)
        (x, w) = np.split(z, [self.prob_num_vars])

        rho = self.rho
        lamb = self.lamb

        eval = self.eval
        obj = eval.obj(x)
        cons = eval.cons(x)
        aug_obj = obj + (rho / 2) * np.dot(cons, cons)

        xdiff = x - self.iterate.x

        if self.rescaled:
            wdiff = w - math.sqrt(lamb) * self.iterate.y
            aug_obj += (lamb / 2) * np.dot(xdiff, xdiff) + (1 / 2) * np.dot(
                wdiff, wdiff
            )
        else:
            wdiff = w - self.iterate.y
            aug_obj += (lamb / 2) * (np.dot(xdiff, xdiff) + np.dot(wdiff, wdiff))

        return float(aug_obj)

    @map_cython_exception
    def gradient(self, z):
        assert z.shape == (self.num_vars,)
        (x, w) = np.split(z, [self.prob_num_vars])

        eval = self.eval
        obj_grad = eval.obj_grad(x)
        cons_jac = eval.cons_jac(x)
        cons = eval.cons(x)

        rho = self.rho
        lamb = self.lamb

        cons_prod = cons_jac.T.dot(cons)

        gradx = obj_grad + rho * cons_prod + lamb * (x - self.iterate.x)

        if self.rescaled:
            gradw = w - math.sqrt(lamb) * self.iterate.y
        else:
            gradw = lamb * (w - self.iterate.y)
        grad = np.concatenate([gradx, gradw])

        assert grad.shape == (self.num_vars,)

        return astype(grad, np.float64)

    @map_cython_exception
    def constraints(self, z):
        assert z.shape == (self.num_vars,)
        (x, w) = np.split(z, [self.prob_num_vars])

        eval = self.eval
        prob_cons = eval.cons(x)
        lamb = self.lamb

        if self.rescaled:
            cons = prob_cons + math.sqrt(lamb) * w
        else:
            cons = prob_cons + lamb * w

        assert cons.shape == (self.num_cons,)

        return astype(cons, np.float64)

    @map_cython_exception
    def jacobian(self, z):
        assert z.shape == (self.num_vars,)
        (x, w) = np.split(z, [self.prob_num_vars])

        eval = self.eval
        problem = self.problem
        jac_x = eval.cons_jac(x).tocoo()
        data_x = jac_x.data

        if self.rescaled:
            data_w = np.full((problem.num_cons,), math.sqrt(self.lamb))
        else:
            data_w = np.full((problem.num_cons,), self.lamb)

        assert self._jac_nnz is not None

        data = np.concatenate([data_x, data_w])

        assert data.shape == (self._jac_nnz,)

        return astype(data, np.float64)

    def jacobianstructure(self):
        problem = self.problem
        iterate = self.iterate
        x = iterate.x

        eval = self.eval
        jac_x = eval.cons_jac(x).tocoo()

        (r_x, c_x) = (jac_x.row, jac_x.col)
        (r_w, c_w) = np.diag_indices(problem.num_cons)
        r_w = np.copy(r_w)
        c_w = np.copy(c_w)

        c_w += problem.num_vars

        rows = np.concatenate([r_x, r_w])
        cols = np.concatenate([c_x, c_w])

        self._jac_nnz = rows.size

        assert rows.shape == cols.shape
        assert rows.ndim == 1

        assert (rows >= 0).all()
        assert (cols >= 0).all()
        assert (rows < self.num_cons).all()
        assert (cols < self.num_vars).all()

        assert rows.dtype == np.int64
        assert cols.dtype == np.int64

        return rows, cols

    def hessian(self, x, lagrange, obj_factor):
        raise NotImplementedError()

    def set_options(self, timer):
        import logging

        logging.getLogger("cyipopt").setLevel(logging.WARNING)
        self.add_option("print_level", 0)
        # self.add_option("derivative_test", "first-order")
        self.add_option("hessian_approximation", "limited-memory")

        remaining = timer.remaining()

        if remaining <= 0.0:
            raise StepSolverError("Time limit reached")
        elif np.isfinite(remaining):
            self.add_option("max_cpu_time", remaining)

    def solve(self, timer):
        iterate = self.iterate

        self.set_options(timer)
        x0 = iterate.x
        y0 = iterate.y
        z0 = np.concatenate([x0, y0])

        # Solve using Ipopt
        try:
            z, info = super().solve(z0)
            assert np.isfinite(z).all()
        except EvalError:
            raise StepSolverError("Failed to evaluate subproblem")

        if info["status"] not in [0, 1]:
            raise StepSolverError("Ipopt failed to solve subproblem")

        x = z[: self.prob_num_vars]
        y = info["mult_g"]

        return (x, y)


class OptimizingController(StepController):
    def step(self, iterate, rho, dt, display, timer) -> StepControlResult:

        problem = self.problem
        params = self.params

        lamb = 1.0 / dt
        implicit_problem = ImplicitProblem(iterate, lamb, rho)
        (x, y) = implicit_problem.solve(timer=timer)

        next_iterate = Iterate(problem, params, x, y, iterate.eval)

        return StepControlResult(
            next_iterate, 0.5 * lamb, active_set=None, rcond=None, accepted=True
        )



import abc
from typing import List, Tuple

import numpy as np

from pygradflow.iterate import Iterate
from pygradflow.log import logger
from pygradflow.params import Params, PenaltyUpdate
from pygradflow.problem import Problem


class PenaltyResult:
    def __init__(self, next_rho, accept) -> None:
        self.next_rho = next_rho
        self.accept = accept

    @staticmethod
    def accept_with_penalty(next_rho):
        return PenaltyResult(next_rho, True)

    @staticmethod
    def reject_with_penalty(next_rho):
        return PenaltyResult(next_rho, False)


class PenaltyStrategy(abc.ABC):
    def __init__(self, problem: Problem, params: Params) -> None:
        self.problem = problem
        self.params = params

    def initial(self, iterate) -> float:
        return self.params.rho

    def update(self, prev_iterate, next_iterate) -> PenaltyResult:
        raise NotImplementedError()


class ConstantPenalty(PenaltyStrategy):
    def __init__(self, problem: Problem, params: Params) -> None:
        super().__init__(problem, params)

    def update(self, prev_iterate: Iterate, next_iterate: Iterate) -> PenaltyResult:
        return PenaltyResult.accept_with_penalty(self.params.rho)


class DualNormUpdate(PenaltyStrategy):
    """
    Increase penalty such that it is within a factor of the inf-norm
    of the dual variables
    """

    def __init__(self, problem: Problem, params: Params) -> None:
        super().__init__(problem, params)

    def initial(self, iterate: Iterate) -> float:
        self.rho = self.params.rho
        return self.rho

    def update(self, prev_iterate: Iterate, next_iterate: Iterate) -> PenaltyResult:
        iterate = next_iterate

        if self.problem.num_cons == 0:
            return PenaltyResult.accept_with_penalty(self.rho)

        ynorm = float(np.linalg.norm(iterate.y, ord=np.inf))

        assert ynorm >= 0.0

        if ynorm >= 10.0 * self.rho:
            next_rho = min(ynorm, 10.0 * self.rho)
            assert next_rho > self.rho
            self.rho = next_rho

        return PenaltyResult.accept_with_penalty(self.rho)


class DualEquilibration(PenaltyStrategy):
    """
    Increase penalty such that the dual the violation is within
    a factor of the product of the dual variables and the constraint values
    """

    def __init__(self, problem: Problem, params: Params):
        super().__init__(problem, params)

    def initial(self, iterate: Iterate) -> float:
        self.rho = self.params.rho
        return self.rho

    def update(self, prev_iterate: Iterate, next_iterate: Iterate) -> PenaltyResult:
        iterate = next_iterate

        cons = iterate.cons

        yprod = abs(np.dot(iterate.y, cons))
        viol = 1.0 / 2.0 * np.dot(cons, cons)

        assert yprod >= 0.0
        assert viol >= 0.0

        if viol == 0.0:
            return PenaltyResult.accept_with_penalty(self.rho)

        target_rho = 0.01 * yprod / viol

        if self.rho < target_rho:
            next_rho = max(self.rho * 10.0, target_rho)
            assert next_rho > self.rho

            self.rho = next_rho

        return PenaltyResult.accept_with_penalty(self.rho)


class ParetoDecrease(PenaltyStrategy):
    """
    Increase penalty such that at least one of objective
    and constraint violation is (weakly) reduced in the direction
    of the curve
    """

    def __init__(self, problem: Problem, params: Params):
        super().__init__(problem, params)

        if problem.var_bounded:
            logger.warning(
                "Pareto decrease penalty update may not work with variable bounds"
            )

    def initial(self, iterate: Iterate) -> float:
        self.rho = self.params.rho
        return self.rho

    def update(self, prev_iterate: Iterate, next_iterate: Iterate) -> PenaltyResult:
        iterate = next_iterate
        params = self.params

        cons = iterate.cons

        viol = 1.0 / 2.0 * np.dot(cons, cons)

        # Don't update if we are already feasible
        if viol <= params.opt_tol:
            return PenaltyResult.accept_with_penalty(self.rho)

        cons_jac = iterate.cons_jac

        infeas_opt_res = cons_jac.T.dot(cons)

        # Cannot find bound if we are locally infeasible
        if np.linalg.norm(infeas_opt_res, ord=np.inf) <= params.local_infeas_tol:
            return PenaltyResult.accept_with_penalty(self.rho)

        obj_bound = np.inf

        obj_grad = iterate.obj_grad
        obj_prod = np.dot(obj_grad, infeas_opt_res)

        cons_dual_prod = cons_jac.T.dot(iterate.y)

        if abs(obj_prod) > 1e-10:
            obj_grad_norm = np.linalg.norm(obj_grad)
            lhs = -(obj_grad_norm + cons_dual_prod.dot(obj_grad))
            obj_bound = lhs / obj_prod

        infeas_res_norm = np.linalg.norm(infeas_opt_res)

        lhs = -np.dot(infeas_opt_res, obj_grad + cons_dual_prod)

        cons_bound = lhs / infeas_res_norm

        bound = min(obj_bound, cons_bound)

        assert np.isfinite(bound)

        next_rho = min(self.rho * 10.0, bound)
        next_rho = max(next_rho, self.rho)

        assert next_rho >= self.rho

        self.rho = next_rho

        return PenaltyResult.accept_with_penalty(self.rho)


class PenaltyFilter(PenaltyStrategy):
    """
    Filter maintaining a set of Pareto optimal points of
    objective and constraint violation. The penalty is
    increased at a point if it is dominated by any
    other point in the filter.

    Note: The running time is linear in the size of the
    filter, this could be improved to be logarithmic
    """

    def __init__(self, problem: Problem, params: Params) -> None:
        super().__init__(problem, params)
        self.entries: List[Tuple[float, float]] = []
        self.rho = self.params.rho

    def filter_insert(self, first, second) -> bool:
        entry = (first, second)

        def dominates(first, second):
            return first[0] <= second[0] and first[1] <= second[1]

        if any(dominates(e, entry) for e in self.entries):
            return False

        self.entries = [e for e in self.entries if not dominates(entry, e)]
        self.entries.append(entry)

        return True

    @abc.abstractmethod
    def iterate_entry(self, iterate):
        raise NotImplementedError()

    def update(self, prev_iterate, next_iterate) -> PenaltyResult:
        next_entry = self.iterate_entry(next_iterate)

        if self.filter_insert(*next_entry):
            return PenaltyResult.accept_with_penalty(self.rho)

        self.rho *= 10.0
        return PenaltyResult.reject_with_penalty(self.rho)


class ObjectivePenaltyFilter(PenaltyFilter):
    def __init__(self, problem: Problem, params: Params) -> None:
        super().__init__(problem, params)

    def iterate_entry(self, iterate):
        obj = iterate.obj
        violation = iterate.cons_violation

        return (obj, violation)


class LagrangianPenaltyFilter(PenaltyFilter):
    def __init__(self, problem: Problem, params: Params) -> None:
        super().__init__(problem, params)

    def iterate_entry(self, iterate):
        rho = self.rho

        lag_x = iterate.aug_lag_deriv_x(rho)
        lag_y = iterate.aug_lag_deriv_y()
        lag_norm_sq = np.dot(lag_x, lag_x) + np.dot(lag_y, lag_y)

        cons = iterate.cons
        cons_val = np.linalg.norm(cons)

        return (lag_norm_sq, cons_val)


def penalty_strategy(problem: Problem, params: Params) -> PenaltyStrategy:
    penalty_update = params.penalty_update

    if penalty_update == PenaltyUpdate.Constant:
        return ConstantPenalty(problem, params)
    elif penalty_update == PenaltyUpdate.DualNorm:
        return DualNormUpdate(problem, params)
    elif penalty_update == PenaltyUpdate.DualEquilibration:
        return DualEquilibration(problem, params)
    elif penalty_update == PenaltyUpdate.ParetoDecrease:
        return ParetoDecrease(problem, params)
    elif penalty_update == PenaltyUpdate.ObjectiveFilter:
        return ObjectivePenaltyFilter(problem, params)
    elif penalty_update == PenaltyUpdate.LagrangianFilter:
        return LagrangianPenaltyFilter(problem, params)

    raise ValueError("Invalid penalty update strategy")



from typing import Callable

import numpy as np
import scipy as sp

from pygradflow.eval import EvalError


class BoxSolverError(Exception):
    pass


# Based on "Projected Newton Methods for Optimization Problems with Simple Constraints"
def solve_box_constrained(
    x0: np.ndarray,
    func: Callable[[np.ndarray], float],
    grad: Callable[[np.ndarray], np.ndarray],
    hess: Callable[[np.ndarray], sp.sparse.spmatrix],
    lb: np.ndarray,
    ub: np.ndarray,
    max_it=1000,
    atol: float = 1e-8,
    rtol: float = 1e-8,
):

    (n,) = x0.shape
    assert lb.shape == (n,)
    assert ub.shape == (n,)

    curr_x = np.clip(x0, lb, ub)

    beta = 0.5
    sigma = 1e-3

    for iteration in range(max_it):
        curr_func = func(curr_x)
        curr_grad = grad(curr_x)

        assert curr_grad.shape == (n,)

        at_lower = np.isclose(curr_x, lb)
        active_lower = np.logical_and(at_lower, curr_grad > 0)

        at_upper = np.isclose(curr_x, ub)
        active_upper = np.logical_and(at_upper, curr_grad < 0)

        residuals = -curr_grad
        residuals[at_lower] = np.maximum(residuals[at_lower], 0)
        residuals[at_upper] = np.minimum(residuals[at_upper], 0)

        residuum = np.linalg.norm(residuals, ord=np.inf)
        grad_norm = np.linalg.norm(curr_grad, ord=np.inf)

        if grad_norm < atol:
            break

        if (residuum < atol) or (residuum / grad_norm) < rtol:
            break

        active = np.logical_or(active_lower, active_upper)
        inactive = np.logical_not(active)

        dir = np.zeros((n,))
        inactive_grad = curr_grad[inactive]

        curr_hess = hess(curr_x)
        assert curr_hess.shape == (n, n)
        inactive_hess = curr_hess.tocsr()[inactive, :][:, inactive]
        dir[inactive] = sp.sparse.linalg.spsolve(inactive_hess, -inactive_grad)

        if np.dot(dir, curr_grad) >= 0:
            raise BoxSolverError("Inactive Hessian not positive definite")

        alpha = 1.0

        eval_errors = False

        for i in range(20):
            next_x = np.clip(curr_x + alpha * dir, lb, ub)

            try:
                next_func = func(next_x)
            except (ArithmeticError, EvalError):
                eval_errors = True
                alpha *= beta
                continue

            rhs = alpha * np.dot(curr_grad[inactive], dir[inactive])

            rhs += np.dot(curr_grad[active], curr_x[active] - next_x[active])

            func_diff = curr_func - next_func

            if func_diff >= sigma * rhs:
                break

            alpha *= beta
        else:
            if eval_errors:
                raise BoxSolverError("Line search failed")
            raise Exception("Line search did not converge")

        curr_x = next_x

    else:
        raise BoxSolverError(f"Did not converge after {max_it} iterations")

    return curr_x



import abc
import functools
from typing import Optional, cast

import numpy as np
import scipy as sp

from pygradflow.implicit_func import StepFunc
from pygradflow.iterate import Iterate
from pygradflow.linear_solver import LinearSolver, LinearSolverError
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.util import norm_mult


class StepResult:
    def __init__(self, orig_iterate, dx, dy, active_set, rcond=None):
        self.orig_iterate = orig_iterate
        self.dy = dy
        self.active_set = active_set
        self.rcond = rcond

        self._compute_xn(dx)

    def _compute_xn(self, dx):
        iterate = self.orig_iterate
        problem = iterate.problem

        var_lb = problem.var_lb
        var_ub = problem.var_ub

        xn = iterate.x - dx
        dx = np.copy(dx)

        at_lb = xn < var_lb
        xn[at_lb] = var_lb[at_lb]
        dx[at_lb] = iterate.x[at_lb] - var_lb[at_lb]

        at_ub = xn > var_ub
        xn[at_ub] = var_ub[at_ub]
        dx[at_ub] = iterate.x[at_ub] - var_ub[at_ub]

        self.dx = dx
        self.xn = xn

    @functools.cached_property
    def iterate(self):
        iterate = self.orig_iterate

        xn = self.xn.astype(iterate.x.dtype, copy=False)
        yn = iterate.y - self.dy

        return Iterate(
            iterate.problem,
            iterate.params,
            xn,
            yn,
            iterate.eval,
        )

    @functools.cached_property
    def diff(self):
        return norm_mult(self.dx, self.dy)


class StepSolver(abc.ABC):
    def __init__(self, problem: Problem, params: Params) -> None:
        self.problem = problem
        self.params = params
        self.n = problem.num_vars
        self.m = problem.num_cons

        self._active_set: Optional[np.ndarray] = None
        self._jac: Optional[sp.sparse.spmatrix] = None
        self._hess: Optional[sp.sparse.spmatrix] = None

        self.solver: Optional[LinearSolver] = None

    @property
    def active_set(self) -> np.ndarray:
        assert self._active_set is not None
        return cast(np.ndarray, self._active_set)

    @property
    def jac(self) -> sp.sparse.spmatrix:
        assert self._jac is not None
        return cast(sp.sparse.spmatrix, self._jac)

    @property
    def hess(self) -> sp.sparse.spmatrix:
        assert self._hess is not None
        return cast(sp.sparse.spmatrix, self._hess)

    def linear_solver(self, mat: sp.sparse.spmatrix) -> LinearSolver:
        from pygradflow.linear_solver import linear_solver

        solver_type = self.params.linear_solver_type
        return linear_solver(mat, solver_type)

    def estimate_rcond(
        self, mat: sp.sparse.spmatrix, solver: LinearSolver
    ) -> Optional[float]:
        from pygradflow.step.cond_estimate import ConditionEstimator

        estimator = ConditionEstimator(mat, solver, self.params)
        rcond = None

        try:
            rcond = estimator.estimate_rcond()
        except LinearSolverError:
            pass

        return rcond

    @abc.abstractmethod
    def update_active_set(self, active_set: np.ndarray):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def func(self) -> StepFunc:
        raise NotImplementedError()

    @abc.abstractmethod
    def update_derivs(self, iterate: Iterate):
        raise NotImplementedError()

    @abc.abstractmethod
    def solve(self, iterate: Iterate) -> StepResult:
        raise NotImplementedError()



from typing import Callable, Union, cast

import numpy as np
import scipy as sp

from pygradflow.params import Params


class DerivError(ValueError):
    def __init__(self, expected_value, actual_value, col_index, atol) -> None:
        self.expected_value = expected_value
        self.actual_value = actual_value
        self.atol = atol

        self.invalid_deriv = np.isclose(
            self.expected_value, self.actual_value, atol=self.atol
        )

        self.invalid_deriv = np.logical_not(self.invalid_deriv)
        self.invalid_indices = np.where(self.invalid_deriv)[0]

        self.deriv_diffs = np.abs(self.expected_value - self.actual_value)
        self.max_deriv_diff = self.deriv_diffs.max()

        (num_rows, _) = self.deriv_diffs.shape
        self.col_index = col_index

    def __str__(self):
        num_invalid_indices = self.invalid_indices.shape[0]

        message = (
            f"Expected and actual (findiff) derivative "
            f"differ at {num_invalid_indices} indices:\n"
        )

        diffs = []

        for invalid_index in self.invalid_indices:
            expected = self.expected_value[invalid_index].item()
            actual = self.actual_value[invalid_index].item()

            diffs.append(f"{invalid_index} | {expected} != {actual}")

        diffs = "\n".join(diffs)

        return message + diffs


def deriv_check(
    f: Callable,
    xval: np.ndarray,
    dval: Union[np.ndarray, sp.sparse.spmatrix],
    params: Params,
) -> None:
    (n,) = xval.shape

    fval = f(xval)
    fval = np.atleast_1d(fval)

    (m,) = fval.shape

    sparse_dval = sp.sparse.issparse(dval)

    eps = params.deriv_pert

    dsparse = False

    if sparse_dval:
        dval = cast(sp.sparse.spmatrix, dval).tocsc()
        dsparse = True
    else:
        dval = np.atleast_2d(dval)

    assert dval.shape == (m, n)

    xtest = np.copy(xval)

    for i in range(n):
        xtest[i] += eps

        testval = f(xtest)

        apx_dval = (testval - fval) / eps

        if apx_dval.ndim == 1:
            apx_dval = apx_dval[:, np.newaxis]

        darray = dval[:, i]

        if dsparse:
            darray = darray.toarray()

        darray = np.atleast_2d(darray)

        assert darray.shape == apx_dval.shape

        if not np.allclose(darray, apx_dval, atol=params.deriv_tol):
            raise DerivError(darray, apx_dval, i, params.deriv_tol)

        xtest[i] -= eps



from pygradflow.iterate import Iterate
from pygradflow.params import Params, StepSolverType
from pygradflow.problem import Problem

from .asymmetric_step_solver import AsymmetricStepSolver
from .extended_step_solver import ExtendedStepSolver
from .standard_step_solver import StandardStepSolver
from .step_solver import StepSolver
from .symmetric_step_solver import SymmetricStepSolver


def step_solver(
    problem: Problem, params: Params, iterate: Iterate, dt: float, rho: float
) -> StepSolver:
    assert dt > 0.0
    assert rho > 0.0

    if params.step_solver is not None:
        return params.step_solver(problem, params, iterate, dt, rho)

    step_solver_type = params.step_solver_type

    if step_solver_type == StepSolverType.Standard:
        return StandardStepSolver(problem, params, iterate, dt, rho)
    elif step_solver_type == StepSolverType.Extended:
        return ExtendedStepSolver(problem, params, iterate, dt, rho)
    elif step_solver_type == StepSolverType.Symmetric:
        return SymmetricStepSolver(problem, params, iterate, dt, rho)
    else:
        assert step_solver_type == StepSolverType.Asymmetric
        return AsymmetricStepSolver(problem, params, iterate, dt, rho)



import scipy as sp

from pygradflow.log import logger

from .linear_solver import LinearSolver, LinearSolverError


class LUSolver(LinearSolver):
    def __init__(self, mat: sp.sparse.spmatrix, symmetric=False) -> None:
        super().__init__(mat, symmetric=symmetric)

        self.mat = mat
        try:
            self.solver = sp.sparse.linalg.splu(mat)
        except RuntimeError as err:
            logger.warn("LU decomposition failed: %s", err)
            raise LinearSolverError("LU decomposition failed")

    def solve(self, rhs, trans=False, initial_sol=None):
        trans_str = "T" if trans else "N"
        return self.solver.solve(rhs, trans=trans_str)



import copy

import numpy as np

from pygradflow.implicit_func import ImplicitFunc
from pygradflow.iterate import Iterate
from pygradflow.linear_solver import LinearSolverError
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.step.step_solver_error import StepSolverError

from .step_solver import StepResult, StepSolver


class StandardStepSolver(StepSolver):
    """
    Standard Newton step solver. Computes the step based on the
    value and derivative of the implicit function.
    """

    def __init__(
        self,
        problem: Problem,
        params: Params,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
    ) -> None:
        super().__init__(problem, params)

        self.orig_iterate = orig_iterate
        self._func = ImplicitFunc(problem, orig_iterate, dt)

        self.rho = rho

    @property
    def func(self) -> ImplicitFunc:
        return self._func

    def _compute_deriv(self) -> None:
        assert self.active_set is not None
        assert self.jac is not None
        assert self.hess is not None

        self.deriv = self.func.deriv(self.jac, self.hess, self.active_set)

    def _reset_deriv(self) -> None:
        self.deriv = None
        self.solver = None

    def update_derivs(self, iterate: Iterate) -> None:
        self._jac = copy.copy(iterate.aug_lag_deriv_xy())
        self._hess = copy.copy(iterate.aug_lag_deriv_xx(self.rho))
        self._reset_deriv()

    def update_active_set(self, active_set: np.ndarray) -> None:
        self._active_set = copy.copy(active_set)
        self._reset_deriv()

    def solve(self, iterate: Iterate) -> StepResult:
        if self.deriv is None:
            self._compute_deriv()

        rhs = self.func.value_at(iterate, self.rho, self.active_set)

        try:
            if self.solver is None:
                self.solver = self.linear_solver(self.deriv)

            sol = self.solver.solve(rhs)
        except LinearSolverError as e:
            raise StepSolverError from e

        n = self.n
        m = self.m

        assert sol.shape == (n + m,)

        dx = sol[:n]
        dy = sol[n:]

        params = self.params

        rcond = None

        if params.report_rcond:
            try:
                rcond = self.estimate_rcond(self.deriv, self.solver)
            except LinearSolverError:
                pass

        return StepResult(iterate, dx, dy, self.active_set, rcond)



from collections import defaultdict
from enum import Enum, auto


class CallbackType(Enum):
    ComputedStep = auto()


class CallbackHandle:
    def __init__(self, callback_type: CallbackType, callback):
        self.callback_type = callback_type
        self.callback = callback

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)

    def __repr__(self) -> str:
        return "CallbackHandle(callback_type={0})".format(self.callback_type)


class Callbacks:
    def __init__(self):
        self._callbacks = defaultdict(lambda: list())

    def register(self, callback_type: CallbackType, callback):
        handle = CallbackHandle(callback_type, callback)
        self._callbacks[callback_type].append(handle)
        return handle

    def unregister(self, handle):
        self._callbacks[handle.callback_type].remove(handle)

    def __call__(self, callback_type, *args, **kwargs):
        for handle in self._callbacks[callback_type]:
            handle(*args, **kwargs)



import scipy as sp

from pygradflow.params import LinearSolverType

from .linear_solver import LinearSolver, LinearSolverError


def linear_solver(
    mat: sp.sparse.spmatrix, solver_type: LinearSolverType, symmetric=False
) -> LinearSolver:
    if solver_type == LinearSolverType.LU:
        from .lu_solver import LUSolver

        return LUSolver(mat, symmetric=symmetric)
    elif solver_type == LinearSolverType.MINRES:
        from .minres_solver import MINRESSolver

        return MINRESSolver(mat, symmetric=symmetric)
    elif solver_type == LinearSolverType.Cholesky:
        from .cholesky_solver import CholeskySolver

        return CholeskySolver(mat, symmetric=symmetric)
    elif solver_type == LinearSolverType.MA57:
        from .ma57_solver import MA57Solver

        return MA57Solver(mat, symmetric=symmetric)
    elif solver_type == LinearSolverType.MUMPS:
        from .mumps_solver import MUMPSSolver

        return MUMPSSolver(mat, symmetric=symmetric)
    elif solver_type == LinearSolverType.SSIDS:
        from .ssids_solver import SSIDSSolver

        return SSIDSSolver(mat, symmetric=symmetric)
    else:
        from .gmres_solver import GMRESSolver

        assert solver_type == LinearSolverType.GMRES
        return GMRESSolver(mat, symmetric=symmetric)


__all__ = ["linear_solver", "LinearSolverError", "LinearSolver"]



from abc import ABC, abstractmethod

from pygradflow.solver import Solver


class Instance(ABC):
    def __init__(self, name, num_vars, num_cons):
        self.name = name
        self.num_vars = num_vars
        self.num_cons = num_cons

    @property
    def size(self):
        return self.num_vars + self.num_cons

    def solve(self, params):
        problem = self.problem()
        solver = Solver(problem, params)
        return solver.solve(self.x0(), self.y0())

    @abstractmethod
    def problem(self):
        raise NotImplementedError()

    @abstractmethod
    def x0(self):
        raise NotImplementedError()

    def y0(self):
        return 0.0



import warnings

import mumps
import numpy as np
import scipy as sp

from .linear_solver import LinearSolver, LinearSolverError


class MUMPSSolver(LinearSolver):
    def __init__(self, mat: sp.sparse.spmatrix, symmetric=False) -> None:
        super().__init__(mat, symmetric=symmetric)

        sym = 2 if symmetric else 0

        self.ctx = mumps.DMumpsContext(sym=sym, par=1)

        # Only output error messages
        self.ctx.set_icntl(4, 1)

        self.ctx.set_icntl(13, 1)

        self.mat = mat

        if mat.format != "coo":
            warnings.warn(
                "Converting matrix to COO format", sp.sparse.SparseEfficiencyWarning
            )
            self.mat = mat.tocoo()

        rows = self.mat.row
        cols = self.mat.col
        data = self.mat.data

        if symmetric:
            filter = rows >= cols
            rows = rows[filter]
            cols = cols[filter]
            data = data[filter]

        if data.dtype != np.float64:
            warnings.warn(
                "Converting matrix data to float64", sp.sparse.SparseEfficiencyWarning
            )
            data = data.astype(np.float64)

        self.ctx.set_shape(self.mat.shape[0])
        self.ctx.set_centralized_assembled(rows + 1, cols + 1, data)

        # Analysis
        self.ctx.run(job=1)

        # Factorization
        self.ctx.run(job=2)

    def solve(self, rhs, trans=False, initial_sol=None):
        sol = np.copy(rhs)

        if sol.dtype != np.float64:
            warnings.warn(
                "Converting rhs to float64", sp.sparse.SparseEfficiencyWarning
            )
            sol = sol.astype(np.float64)

        if not self.symmetric:
            if trans:
                self.ctx.set_icntl(9, 0)
            else:
                self.ctx.set_icntl(9, 1)

        self.ctx.set_rhs(sol)

        try:
            # Solution
            self.ctx.run(job=3)
        except RuntimeError as e:
            raise LinearSolverError from e

        return sol.astype(rhs.dtype)

    def num_neg_eigvals(self):
        return self.ctx.id.infog[11]

    def rcond(self):
        return None



import os

import pyqplib

from pygradflow.solver import Problem

from .instance import Instance
from .runner import Runner


class QPLIBProblem(Problem):
    def __init__(self, problem):
        self.problem = problem

        super().__init__(
            problem.var_lb,
            problem.var_ub,
            cons_lb=problem.cons_lb,
            cons_ub=problem.cons_ub,
        )

    def obj(self, x):
        return self.problem.obj_val(x)

    def obj_grad(self, x):
        return self.problem.obj_grad(x)

    def cons(self, x):
        return self.problem.cons_val(x)

    def cons_jac(self, x):
        return self.problem.cons_jac(x)

    def lag_hess(self, x, y):
        return self.problem.lag_hess(x, y)

    def x0(self):
        return self.problem.x0

    def y0(self):
        return self.problem.y0


class QPLIBInstance(Instance):
    def __init__(self, description):
        self.description = description

        super().__init__(description.name, description.num_vars, description.num_cons)

    @property
    def filename(self):
        return self.description.filename

    def problem(self):
        qproblem = pyqplib.read_problem(self.filename)
        return QPLIBProblem(qproblem)

    def x0(self):
        return self.problem().x0()

    def y0(self):
        return self.problem().y0()


class QPLIBRunner(Runner):
    def __init__(self):
        super().__init__(name="qplib")

    def get_instances(self, args):
        instances = []
        for file in os.listdir(args.directory):
            is_qplib = file.endswith(".qplib") or file.endswith(".qplib.zip")

            if is_qplib:
                filename = os.path.join(args.directory, file)
                description = pyqplib.read_description(filename)
                instances.append(QPLIBInstance(description))

        return instances

    def parser(self):
        parser = super().parser()
        parser.add_argument("directory")
        return parser


if __name__ == "__main__":
    QPLIBRunner().main()



import numpy as np


class ActiveSet:
    def __init__(self, iterate):
        active_tol = iterate.params.active_tol

        lb = iterate.problem.var_lb
        ub = iterate.problem.var_ub

        at_lower = np.absolute(iterate.x - lb) <= active_tol
        at_upper = np.absolute(ub - iterate.x) <= active_tol

        violated_lower = lb - iterate.x > active_tol
        violated_upper = iterate.x - ub > active_tol

        self.violated = np.logical_or(violated_lower, violated_upper)

        self.at_either = np.logical_or(at_lower, at_upper)

        self.at_both = np.logical_and(at_upper, at_lower)
        not_at_both = np.logical_not(self.at_both)

        self.at_lower = np.logical_and(at_lower, not_at_both)
        self.at_upper = np.logical_and(at_upper, not_at_both)

    @property
    def satisfied(self):
        return np.logical_not(self.violated)



import abc
import math
from enum import Enum, auto

import numpy as np
import scipy as sp

from pygradflow.log import logger


def astype(array, dtype):
    if array.dtype == dtype:
        return array
    else:
        return array.astype(dtype)


class EvalError(ValueError):
    def __init__(self, msg, x):
        self.x = x
        super().__init__(msg)


def warn_once(*args):
    has_warned = [False]

    def warn():
        if not has_warned[0]:
            logger.warning(*args)
            has_warned[0] = True

    return warn


warn_hessian_pattern = warn_once("Unsymmetric Hessian pattern")
warn_hessian_values = warn_once("Hessian not numerically symmetric")


class Component(Enum):
    Obj = auto()
    ObjGrad = auto()
    Cons = auto()
    ConsJac = auto()
    LagHess = auto()

    def name(self):
        return {
            Component.Obj: "Objective",
            Component.ObjGrad: "Objective Gradient",
            Component.Cons: "Constraints",
            Component.ConsJac: "Constraint Jacobian",
            Component.LagHess: "Lagrangian Hessian",
        }[self]


class Evaluator(abc.ABC):
    def __init__(self, problem, params):
        self.problem = problem
        self.dtype = params.dtype

        self.reset_num_evals()

    def reset_num_evals(self):
        self.num_evals = {comp: 0 for comp in Component}

    def obj(self, x: np.ndarray) -> float:
        self.num_evals[Component.Obj] += 1
        return self._eval_obj(x)

    def obj_grad(self, x: np.ndarray) -> np.ndarray:
        self.num_evals[Component.ObjGrad] += 1
        return self._eval_obj_grad(x)

    def cons(self, x: np.ndarray) -> np.ndarray:
        self.num_evals[Component.Cons] += 1
        return self._eval_cons(x)

    def cons_jac(self, x: np.ndarray) -> sp.sparse.spmatrix:
        self.num_evals[Component.ConsJac] += 1
        return self._eval_cons_jac(x)

    def lag_hess(self, x: np.ndarray, lag: np.ndarray) -> sp.sparse.spmatrix:
        self.num_evals[Component.LagHess] += 1
        return self._eval_lag_hess(x, lag)

    @abc.abstractmethod
    def _eval_obj(self, x: np.ndarray) -> float:
        raise NotImplementedError()

    @abc.abstractmethod
    def _eval_obj_grad(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    @abc.abstractmethod
    def _eval_cons(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    @abc.abstractmethod
    def _eval_cons_jac(self, x: np.ndarray) -> sp.sparse.spmatrix:
        raise NotImplementedError()

    @abc.abstractmethod
    def _eval_lag_hess(self, x: np.ndarray, lag: np.ndarray) -> sp.sparse.spmatrix:
        raise NotImplementedError()


class SimpleEvaluator(Evaluator):
    def _eval_obj(self, x: np.ndarray) -> float:
        return self.problem.obj(x)

    def _eval_obj_grad(self, x: np.ndarray) -> np.ndarray:
        return astype(self.problem.obj_grad(x), self.dtype)

    def _eval_cons(self, x: np.ndarray) -> np.ndarray:
        if self.problem.num_cons == 0:
            return np.array([], dtype=self.dtype)

        return astype(self.problem.cons(x), self.dtype)

    def _eval_cons_jac(self, x: np.ndarray) -> sp.sparse.spmatrix:
        if self.problem.num_cons == 0:
            return sp.sparse.csr_matrix((0, self.problem.num_vars), dtype=self.dtype)

        return astype(self.problem.cons_jac(x), self.dtype)

    def _eval_lag_hess(self, x: np.ndarray, lag: np.ndarray) -> sp.sparse.spmatrix:
        return astype(self.problem.lag_hess(x, lag), self.dtype)


class ValidatingEvaluator(Evaluator):
    def __init__(self, problem, params):
        super().__init__(problem, params)
        self.num_vars = problem.num_vars
        self.num_cons = problem.num_cons

    def _eval_obj(self, x: np.ndarray) -> float:
        obj = self.problem.obj(x)

        if not math.isfinite(obj):
            raise EvalError("Infinite objective", x)

        return obj

    def _eval_obj_grad(self, x: np.ndarray) -> np.ndarray:
        grad = self.problem.obj_grad(x)

        if grad.shape != (self.num_vars,):
            raise EvalError("Invalid shape of gradient", x)

        if not np.isfinite(grad).all():
            raise EvalError("Non-finite gradient", x)

        return astype(grad, self.dtype)

    def _eval_cons(self, x: np.ndarray) -> np.ndarray:
        if self.num_cons == 0:
            return np.array([], dtype=self.dtype)

        cons = self.problem.cons(x)

        if cons.shape != (self.num_cons,):
            raise EvalError("Invalid shape of constraints", x)

        if not np.isfinite(cons).all():
            raise EvalError("Non-finite constraints", x)

        return astype(cons, self.dtype)

    def _eval_cons_jac(self, x: np.ndarray) -> sp.sparse.spmatrix:
        if self.num_cons == 0:
            return sp.sparse.csr_matrix((0, self.num_vars), dtype=self.dtype)

        cons_jac = self.problem.cons_jac(x)

        if cons_jac.shape != (self.num_cons, self.num_vars):
            raise EvalError("Invalid shape of Jacobian", x)

        if not np.isfinite(cons_jac.data).all():
            raise EvalError("Non-finite Jacobian", x)

        return astype(cons_jac, self.dtype)

    def _eval_lag_hess(self, x: np.ndarray, lag: np.ndarray) -> sp.sparse.spmatrix:
        lag_hess = self.problem.lag_hess(x, lag)

        if lag_hess.shape != (self.num_vars, self.num_vars):
            raise EvalError("Invalid shape of Hessian", x)

        if not np.isfinite(lag_hess.data).all():
            raise EvalError("Non-finite Hessian", x)

        coo_hess = lag_hess.tocoo()
        coo_hess_T = coo_hess.T

        orig_pattern = set(zip(coo_hess.row, coo_hess.col))
        trans_pattern = set(zip(coo_hess.col, coo_hess.row))

        surplus_orig = orig_pattern - trans_pattern
        surplus_trans = trans_pattern - orig_pattern

        same_pattern = not (surplus_orig or surplus_trans)

        if not same_pattern:
            warn_hessian_pattern()
        else:
            coo_data = coo_hess.data
            coo_T_data = coo_hess_T.data
            if not np.allclose(coo_data, coo_T_data):
                warn_hessian_values()

        return astype(lag_hess, self.dtype)


def create_evaluator(problem, params):
    if params.validate_input:
        return ValidatingEvaluator(problem, params)
    else:
        return SimpleEvaluator(problem, params)



import os

import mpspy
import numpy as np

from pygradflow.problem import Problem
from pygradflow.util import sparse_zero

from .instance import Instance
from .runner import Runner


class LinearProblem(Problem):
    def __init__(self, instance):
        self.instance = instance

        super().__init__(
            var_lb=instance.var_lb,
            var_ub=instance.var_ub,
            cons_lb=instance.cons_lb,
            cons_ub=instance.cons_ub,
        )

    def obj(self, x):
        instance = self.instance
        return np.dot(instance.obj, x)

    def obj_grad(self, x):
        instance = self.instance
        return instance.obj

    def cons(self, x):
        instance = self.instance
        coeffs = instance.coeffs
        return coeffs.dot(x)

    def cons_jac(self, x):
        instance = self.instance
        return instance.coeffs

    def lag_hess(self, x, y):
        num_vars = self.num_vars
        return sparse_zero(shape=(num_vars, num_vars))


class MPSInstance(Instance):
    def __init__(self, instance):
        self.instance = instance

        num_vars = instance.num_vars
        num_cons = instance.num_cons
        name = instance.name

        super().__init__(name, num_vars, num_cons)

    def problem(self):
        return LinearProblem(self.instance)

    def x0(self):
        instance = self.instance
        num_vars = instance.num_vars

        x0 = np.zeros((num_vars,))

        return np.clip(x0, instance.var_lb, instance.var_ub)


class MPSRunner(Runner):
    def __init__(self):
        super().__init__(name="mps")

    def get_instances(self, args):
        instances = []
        for file in os.listdir(args.directory):
            is_qplib = file.endswith(".mps") or file.endswith(".mps.gz")

            if is_qplib:
                filename = os.path.join(args.directory, file)
                description = mpspy.read_mps(filename)
                instances.append(MPSInstance(description))

        return instances

    def parser(self):
        parser = super().parser()
        parser.add_argument("directory")
        return parser


if __name__ == "__main__":
    MPSRunner().main()



import numpy as np

from pygradflow.controller import ControllerSettings, LogController
from pygradflow.implicit_func import ImplicitFunc
from pygradflow.log import logger
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.step.newton_control import NewtonController
from pygradflow.step.step_control import StepControlResult


class ResiduumRatioController(NewtonController):
    def __init__(self, problem: Problem, params: Params) -> None:
        settings = ControllerSettings.from_params(params)
        self.controller = LogController(settings, params.theta_ref)
        super().__init__(problem, params)

    def step(self, iterate, rho, dt, display, timer):
        assert dt > 0.0
        lamb = 1.0 / dt

        problem = self.problem
        params = self.params

        next_steps = self.newton_steps(iterate, rho, dt)

        func = ImplicitFunc(problem, iterate, dt)

        mid_step = next(next_steps)
        mid_iterate = mid_step.iterate

        mid_norm = np.linalg.norm(func.value_at(mid_iterate, rho))

        self.display_step(0, mid_step)

        if mid_norm <= params.newton_tol:
            lamb_n = max(lamb * params.lamb_red, params.lamb_min)
            logger.debug("Newton converged during first iteration, lamb_n = %f", lamb_n)
            return StepControlResult.from_step_result(mid_step, lamb_n, True)

        orig_norm = np.linalg.norm(func.value_at(iterate, rho))

        theta = mid_norm / orig_norm
        accepted = theta <= params.theta_max

        if accepted:
            lamb_mod = self.controller.update(theta)
            lamb_n = max(params.lamb_min, lamb / lamb_mod)
        else:
            lamb_n = lamb * params.lamb_inc
            if self.controller.error_sum > 0.0:
                self.controller.reset()

        logger.debug(
            "StepController: theta: %e, accepted: %s, lamb: %e, lamb_n: %e",
            theta,
            accepted,
            lamb,
            lamb_n,
        )

        self.lamb = lamb_n
        return StepControlResult.from_step_result(mid_step, lamb_n, accepted)



import functools
from typing import Optional

import numpy as np
import scipy as sp

from pygradflow.active_set import ActiveSet
from pygradflow.eval import Evaluator, SimpleEvaluator
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.util import norm_mult


def _read_only(a):
    a.flags.writeable = False
    return a


class Iterate:
    def __init__(
        self,
        problem: Problem,
        params: Params,
        x: np.ndarray,
        y: np.ndarray,
        eval: Optional[Evaluator] = None,
    ):
        assert x.shape == (problem.num_vars,)
        assert y.shape == (problem.num_cons,)
        self.x = _read_only(np.copy(x))
        self.y = _read_only(np.copy(y))
        self.params = params

        if eval:
            self.eval = eval
        else:
            self.eval = SimpleEvaluator(problem, params)
        self.problem = problem

    @property
    def z(self):
        return np.concatenate((self.x, self.y))

    def copy(self) -> "Iterate":
        return Iterate(
            self.problem, self.params, np.copy(self.x), np.copy(self.y), self.eval
        )

    def clipped(self) -> "Iterate":
        var_lb = self.problem.var_lb
        var_ub = self.problem.var_ub
        x = self.x
        if np.all(x >= var_lb) and np.all(x <= var_ub):
            return self
        xclip = np.empty_like(x)
        np.clip(x, var_lb, var_ub, out=xclip)
        return Iterate(self.problem, self.params, xclip, self.y, self.eval)

    @functools.cached_property
    def obj(self) -> float:
        return self.eval.obj(self.x)

    @functools.cached_property
    def obj_grad(self) -> np.ndarray:
        return _read_only(self.eval.obj_grad(self.x))

    @functools.cached_property
    def cons(self) -> np.ndarray:
        return _read_only(self.eval.cons(self.x))

    @functools.cached_property
    def cons_jac(self) -> sp.sparse.spmatrix:
        return self.eval.cons_jac(self.x)

    def lag_hess(self, y: np.ndarray) -> sp.sparse.spmatrix:
        return self.eval.lag_hess(self.x, y)

    def aug_lag_violation(self, rho: float) -> float:
        cv = self.cons
        return rho / 2.0 * np.dot(cv, cv)

    def aug_lag_dual(self) -> float:
        cv = self.cons
        return np.dot(cv, self.y)

    def aug_lag(self, rho: float) -> float:
        o = self.obj
        cv = self.cons
        return o + rho / 2.0 * np.dot(cv, cv) + np.dot(cv, self.y)

    def aug_lag_deriv_x(self, rho: float) -> np.ndarray:
        lhs = rho * self.cons + self.y

        return self.obj_grad + self.cons_jac.T.dot(lhs)

    def aug_lag_deriv_y(self) -> np.ndarray:
        return self.cons

    def aug_lag_deriv_xy(self) -> sp.sparse.spmatrix:
        return self.cons_jac

    def aug_lag_deriv_xx(self, rho: float) -> sp.sparse.spmatrix:
        jac = self.cons_jac
        cons = self.cons
        m = self.y + rho * cons

        if rho == 0.0:
            return self.lag_hess(m)
        else:
            return self.lag_hess(m) + rho * np.dot(jac.T, jac)

    def dist(self, other: "Iterate") -> float:
        return norm_mult(self.x - other.x, self.y - other.y)

    def locally_infeasible(self, feas_tol: float, local_infeas_tol: float) -> bool:
        """
        Check if the iterate is locally infeasible. It is
        judged to be locally infeasible if the constraint
        violation is greater than the tolerance and
        optimality conditions for the minimization
        of the constraint violation are (approximately) satisfied.
        """
        if self.cons_violation <= feas_tol:
            return False

        infeas_opt_res = self.cons_jac.T.dot(self.cons)

        at_lower = self.active_set.at_lower
        at_upper = self.active_set.at_upper

        infeas_opt_res[at_lower] = np.minimum(infeas_opt_res[at_lower], 0.0)
        infeas_opt_res[at_upper] = np.maximum(infeas_opt_res[at_upper], 0.0)

        return bool(np.linalg.norm(infeas_opt_res, ord=np.inf) <= local_infeas_tol)

    @functools.cached_property
    def active_set(self) -> ActiveSet:
        return ActiveSet(self)

    @functools.cached_property
    def bounds_dual(self) -> np.ndarray:
        r = -(self.obj_grad + self.cons_jac.T.dot(self.y))
        d = np.zeros_like(self.x)

        active_set = self.active_set

        d[active_set.at_upper] = np.maximum(r[active_set.at_upper], 0.0)
        d[active_set.at_lower] = np.minimum(r[active_set.at_lower], 0.0)
        d[active_set.at_both] = r[active_set.at_both]

        return d

    @functools.cached_property
    def bound_violation(self) -> float:
        lb = self.problem.var_lb
        ub = self.problem.var_ub
        x = self.x

        lower = float(np.linalg.norm(np.maximum(lb - x, 0.0), np.inf))
        upper = float(np.linalg.norm(np.maximum(x - ub, 0.0), np.inf))

        return max(lower, upper)

    @functools.cached_property
    def cons_violation(self) -> float:
        c = self.cons
        if c.size == 0:
            return 0.0
        return float(np.linalg.norm(c, np.inf))

    @functools.cached_property
    def stat_res(self) -> float:
        r = self.obj_grad + self.cons_jac.T.dot(self.y) + self.bounds_dual
        return float(np.linalg.norm(r, np.inf))

    def is_feasible(self, tol):
        return (self.cons_violation <= tol) and (self.bound_violation <= tol)

    @property
    def total_res(self) -> float:
        return max(self.cons_violation, self.bound_violation, self.stat_res)

    def obj_nonlin(self, other: "Iterate") -> float:
        dx = other.x - self.x
        next_obj = self.obj + np.dot(dx, self.obj_grad)
        dx_dot = np.dot(dx, dx)
        if np.isclose(dx_dot, 0.0):
            return 0.0
        return abs(other.obj - next_obj) / dx_dot

    def cons_nonlin(self, other: "Iterate") -> np.ndarray:
        dx = other.x - self.x
        next_cons = self.cons + self.cons_jac.dot(dx)
        dx_dot = np.dot(dx, dx)
        if np.isclose(dx_dot, 0.0):
            problem = self.problem
            return np.zeros((problem.num_cons,))
        return (other.cons - next_cons) / dx_dot

    def check_eval(self):
        # Check for and raise EvalError if any of the
        # evaluations are fail
        self.obj
        self.obj_grad

        if self.problem.num_cons > 0:
            self.cons
            self.cons_jac



import numpy as np
import scipy as sp
from numpy import ndarray


def sparse_zero(shape, format=None):
    zero_mat = sp.sparse.coo_matrix(([], ([], [])), shape)

    if format not in [None, "coo"]:
        return zero_mat.asformat(format)

    return zero_mat


def norm_sq(x: ndarray) -> float:
    return np.dot(x, x)


def norm_mult(*args) -> float:
    value = 0.0
    for arg in args:
        value += norm_sq(arg)

    return np.sqrt(value)


def keep_rows(mat, row_filter):
    (m, _) = mat.shape

    assert row_filter.shape == (m,)
    assert row_filter.dtype == bool

    if row_filter.all():
        return mat

    mat = mat.tocoo()

    rows = mat.row
    cols = mat.col
    data = mat.data

    nnz = mat.nnz
    indices = np.arange(nnz)

    index_filter = row_filter[rows[indices]]

    alive_rows = rows[index_filter]
    alive_cols = cols[index_filter]
    alive_data = data[index_filter]

    cleared_mat = sp.sparse.coo_matrix(
        (alive_data, (alive_rows, alive_cols)), shape=mat.shape
    )

    return cleared_mat



from enum import Enum, auto
from typing import List, Optional, cast

import numpy as np
import scipy as sp

from pygradflow.display import (
    Format,
    StateData,
    integrator_display,
    print_problem_stats,
)
from pygradflow.integration.events import (
    ConvergedResult,
    EventResultType,
    FilterChangedResult,
    PenaltyResult,
    UnboundedResult,
)
from pygradflow.integration.flow import Flow, func_neg, func_pos, lazy_func
from pygradflow.integration.problem_switches import SwitchTrigger, TriggerType
from pygradflow.integration.restricted_flow import RestrictedFlow
from pygradflow.iterate import Iterate
from pygradflow.log import logger
from pygradflow.params import Params
from pygradflow.result import SolverResult
from pygradflow.status import SolverStatus
from pygradflow.timer import Timer
from pygradflow.transform import Transformation


class IntegrationStatus(Enum):
    Converged = auto()
    Unbounded = auto()
    Event = auto()
    Finished = auto()
    Penalty = auto()

    def name(self):
        return {
            IntegrationStatus.Converged: "Converged",
            IntegrationStatus.Unbounded: "Unbounded",
            IntegrationStatus.Event: "Event",
            IntegrationStatus.Finished: "Finished",
            IntegrationStatus.Penalty: "Penalty",
        }[self]


class IntegrationResult:
    def __init__(
        self,
        status,
        dist,
        t_next,
        z_next,
        filter_next,
        num_steps,
        num_func_evals,
        num_jac_evals,
    ):
        self.status = status
        self.dist = dist
        self.num_steps = num_steps
        self.num_func_evals = num_func_evals
        self.num_jac_evals = num_jac_evals
        self.t = t_next
        self.z = z_next
        self.filter = filter_next


class IntegrationSolver:
    def __init__(self, problem, params=Params()):
        self.orig_problem = problem
        self.params = params
        self.path: Optional[List[np.ndarray]] = None

    def create_events(self, result, triggers):
        all_events = []

        for i, trigger in enumerate(triggers):
            t_events = result.t_events[i]
            z_events = result.y_events[i]

            for t, z in zip(t_events, z_events):
                all_events.append(SwitchTrigger(t, z, trigger))

        all_events.sort(key=lambda e: e.time)
        return all_events

    def create_filter(self, z, rho):
        problem = self.problem
        x = z[: problem.num_vars]
        lb = problem.var_lb
        ub = problem.var_ub

        at_lb = Flow.isclose(x, lb)
        at_ub = Flow.isclose(x, ub)
        at_bounds = np.logical_or(at_lb, at_ub)

        dx = self.flow.neg_aug_lag_deriv_x(z, rho)

        active_lower = np.logical_and(at_lb, dx < 0)
        active_upper = np.logical_and(at_ub, dx > 0)

        fixed_indices = np.logical_or(active_lower, active_upper)

        dx_zero = Flow.isclose(dx, 0.0)
        ambiguous = np.logical_and(dx_zero, at_bounds)

        if not (ambiguous.any()):
            return np.logical_not(fixed_indices)

        ddx = self.flow.rhs_deriv_x(z, rho)
        if Flow.isclose(ddx[ambiguous], 0.0).any():
            raise Exception("Degenerate bound")

        ambiguous_lb = np.logical_and(at_lb, dx_zero)
        fixed_indices[ambiguous_lb] = ddx[ambiguous_lb] < 0

        ambiguous_ub = np.logical_and(at_ub, dx_zero)
        fixed_indices[ambiguous_ub] = ddx[ambiguous_ub] > 0

        return np.logical_not(fixed_indices)

    def _check_bounds(self, z):
        problem = self.problem
        x = z[: problem.num_vars]

        lb = problem.var_lb
        ub = problem.var_ub

        between = np.logical_and(lb <= x, x <= ub)
        at_lb = Flow.isclose(x, lb)
        at_ub = Flow.isclose(x, ub)

        valid = np.logical_or(between, np.logical_or(at_lb, at_ub))

        assert valid.all()

    def _check_filter(self, z, filter, rho):
        assert (filter == self.create_filter(z, rho)).all()

    def handle_events(self, events, restricted_flow, rho):
        problem = self.problem
        lb = problem.var_lb
        ub = problem.var_ub
        filter = restricted_flow.filter
        flow = restricted_flow.flow

        logger.debug("Handling %d events", len(events))

        for event in events:
            z_event = event.state
            t_event = event.time

            self._check_bounds(z_event)

            @lazy_func
            def rhs():
                return flow.rhs(z_event, rho)

            @lazy_func
            def rhs_deriv():
                return flow.rhs_deriv_x(z_event, rho)

            if event.type == TriggerType.LB:
                j = event.index
                logger.debug("State %d reached lower bound at time %f", j, event.time)

                assert filter[j]
                assert Flow.isclose(z_event[j], lb[j])
                assert func_neg(rhs, rhs_deriv, j)

                return FilterChangedResult(t_event, z_event, filter, j)

            elif event.type == TriggerType.UB:
                j = event.index
                logger.debug("State %d reached upper bound at time %f", j, event.time)

                assert filter[j]
                assert Flow.isclose(z_event[j], ub[j])
                assert func_pos(rhs, rhs_deriv, j)

                return FilterChangedResult(t_event, z_event, filter, j)

            elif event.type == TriggerType.GRAD_FIXED:
                j = event.index
                logger.debug(
                    "Gradient entry of fixed %d changed sign at time %f", j, event.time
                )

                at_lb = Flow.isclose(z_event[j], lb[j])
                at_ub = Flow.isclose(z_event[j], ub[j])

                assert not (filter[j])

                if at_lb:
                    assert func_pos(rhs, rhs_deriv, j)
                elif at_ub:
                    assert func_neg(rhs, rhs_deriv, j)

                return FilterChangedResult(t_event, z_event, filter, j)

            elif event.type == TriggerType.UNBOUNDED:
                params = self.params
                (x_event, y_event) = self.flow.split_states(z_event)
                iterate_event = Iterate(problem, params, x_event, y_event)

                if iterate_event.is_feasible(params.opt_tol):
                    return UnboundedResult(t_event, z_event)

            elif event.type == TriggerType.PENALTY:
                (x_event, y_event) = self.flow.split_states(z_event)
                return PenaltyResult(t_event, z_event)
            else:
                assert event.type == TriggerType.CONVERGED
                logger.debug("Convergence achieved at time %f", event.time)
                return ConvergedResult(t_event, z_event)

        return None

    def print_result(
        self,
        total_time: float,
        status: SolverStatus,
        iterate: Iterate,
        iterations: int,
        dist_factor: float,
    ) -> None:
        rho = self.rho

        desc = "{:>45s}".format(SolverStatus.description(status))

        status_desc = Format.redgreen(desc, SolverStatus.success(status), bold=True)
        status_name = Format.bold("{:>20s}".format("Status"))

        logger.info("%20s: %45s", status_name, status_desc)
        logger.info("%20s: %45s", "Time", f"{total_time:.2f}s")
        logger.info("%20s: %45d", "Iterations", iterations)

        logger.info("%20s: %45e", "Distance factor", dist_factor)

        logger.info("%20s: %45e", "Objective", iterate.obj)
        logger.info("%20s: %45e", "Aug Lag violation", iterate.aug_lag_violation(rho))
        logger.info("%20s: %45e", "Aug Lag dual", iterate.aug_lag_dual())

        logger.info("%20s: %45e", "Constraint violation", iterate.cons_violation)
        logger.info("%20s: %45e", "Dual violation", iterate.stat_res)

        eval = self.evaluator

        eval_name = Format.bold("{:>20s}".format("Evaluations"))
        logger.info("%20s", eval_name)

        for component, num_evals in eval.num_evals.items():
            name = component.name()
            logger.info("%20s: %45d", name, num_evals)

    def perform_integration(
        self, curr_t, curr_z, curr_filter, rho
    ) -> IntegrationResult:
        # We want to integrate to +oo...
        next_t = curr_t + 1e10

        restricted_flow = RestrictedFlow(self.flow, curr_filter)
        event_triggers = restricted_flow.create_event_triggers(curr_z, rho)

        # Check consistency
        (curr_x, curr_y) = self.flow.split_states(curr_z)
        assert self.flow.is_boxed(curr_x)
        restricted_flow.check_point(curr_z, rho)

        ivp_result = sp.integrate.solve_ivp(
            restricted_flow.rhs_func(rho),
            (curr_t, next_t),
            curr_z,
            method="BDF",
            jac=restricted_flow.rhs_jac_func(rho),
            events=event_triggers,
        )

        assert ivp_result.success, "Failed integration"
        assert (ivp_result.y[:, 0] == curr_z).all()

        events = self.create_events(ivp_result, event_triggers)
        event_result = self.handle_events(events, restricted_flow, rho)

        status = IntegrationStatus.Finished
        next_filter = curr_filter

        if event_result is None:
            next_t = ivp_result.t[-1]
            next_z = ivp_result.y[:, -1]
        else:
            next_z = event_result.z
            next_t = event_result.t

            if event_result.type == EventResultType.CONVERGED:
                logger.debug("Convergence achieved")
                status = IntegrationStatus.Converged
            elif event_result.type == EventResultType.UNBOUNDED:
                logger.debug("Unboundedness detected")
                status = IntegrationStatus.Unbounded
            elif event_result.type == EventResultType.FILTER_CHANGED:
                logger.debug("Filter changed")
                status = IntegrationStatus.Event
                next_filter = event_result.filter
            elif event_result.type == EventResultType.FREE_GRAD_ZERO:
                logger.debug("Free gradient entry %d became zero", event_result.j)
                status = IntegrationStatus.Event
            else:
                assert event_result.type == EventResultType.PENALTY
                logger.debug("Penalty event")
                status = IntegrationStatus.Penalty

        (next_x, next_y) = self.flow.split_states(next_z)

        not_filter = np.logical_not(curr_filter)

        assert (next_x[not_filter] == curr_x[not_filter]).all()

        assert self.flow.is_approx_boxed(next_x)

        next_x = np.clip(next_x, self.problem.var_lb, self.problem.var_ub)
        next_z = np.concatenate((next_x, next_y))

        params = self.params

        if params.collect_path:
            assert self.path is not None
            path = cast(List[np.ndarray], self.path)
            assert (path[-1][:, -1] == curr_z).all()
            assert (path[-1][:, -1] == ivp_result.y[:, 0]).all()
            path.append(ivp_result.y[:, 1:])
            self.path_times.append(ivp_result.t[1:])
            path[-1][:, -1] = next_z

        dist = np.linalg.norm(
            (ivp_result.y[:, 1:] - ivp_result.y[:, :-1]), axis=0
        ).sum()

        return IntegrationResult(
            status,
            dist,
            next_t,
            next_z,
            next_filter,
            num_steps=ivp_result.t.size,
            num_func_evals=ivp_result.nfev,
            num_jac_evals=ivp_result.njev,
        )

    def solve(self, x0: Optional[np.ndarray] = None, y0: Optional[np.ndarray] = None):

        self.transform = Transformation(self.orig_problem, self.params, x0, y0)

        self.problem = self.transform.trans_problem
        self.evaluator = self.transform.evaluator

        problem = self.problem
        params = self.params
        self.flow = Flow(problem, params, self.evaluator)
        self.rho = self.params.rho

        initial_iterate = self.transform.initial_iterate

        self.path = [initial_iterate.z[:, None]]
        self.path_times = [np.array([0.0])]

        print_problem_stats(problem, initial_iterate)

        x_init = initial_iterate.x
        y_init = initial_iterate.y

        curr_z = np.concatenate((x_init, y_init))
        curr_t = 0.0
        curr_filter = self.create_filter(curr_z, self.rho)

        status = None
        iteration = 0
        path_dist = 0.0
        accepted_steps = 0

        timer = Timer(params.time_limit)

        display = integrator_display(problem, params)
        logger.info(display.header)

        while True:
            self._check_filter(curr_z, curr_filter, self.rho)
            self._check_bounds(curr_z)

            restricted_flow = RestrictedFlow(self.flow, curr_filter)
            curr_res = restricted_flow.residuum(curr_z)

            if curr_res <= params.opt_tol:
                logger.debug("Iterate is optimal")
                status = SolverStatus.Optimal
                break

            logger.debug("Iteration %d: value: %s", iteration, curr_z)

            logger.debug(
                "State: %s, filter: %s, grad: %s",
                curr_z,
                curr_filter,
                self.flow.rhs(curr_z, self.rho),
            )

            if timer.reached_time_limit():
                logger.debug("Reached time limit")
                status = SolverStatus.TimeLimit
                break

            (curr_x, curr_y) = self.flow.split_states(curr_z)

            curr_it = Iterate(problem, params, curr_x, curr_y)

            if curr_it.locally_infeasible(params.opt_tol, params.local_infeas_tol):
                logger.debug("Local infeasibility detected")
                status = SolverStatus.LocallyInfeasible
                break

            if (curr_it.obj <= params.obj_lower_limit) and (
                curr_it.is_feasible(params.opt_tol)
            ):
                logger.debug("Unboundedness detected")
                status = SolverStatus.Unbounded
                break

            result = self.perform_integration(curr_t, curr_z, curr_filter, self.rho)
            path_dist += result.dist

            if display.should_display():
                state = StateData()
                curr_x, curr_y = self.flow.split_states(curr_z)
                iterate = curr_it
                state["iterate"] = iterate
                state["filter"] = curr_filter
                state["aug_lag"] = lambda: iterate.aug_lag(self.rho)
                state["obj"] = lambda: iterate.obj
                state["iter"] = iteration + 1
                state["num_func_evals"] = result.num_func_evals
                state["num_jac_evals"] = result.num_jac_evals
                state["num_steps"] = result.num_steps
                state["dt"] = result.t - curr_t
                state["step_type"] = result.status

                logger.info(display.row(state))

            iteration += 1

            curr_z = result.z
            curr_t = result.t
            curr_filter = result.filter

            if result.status == IntegrationStatus.Converged:
                status = SolverStatus.Optimal
                break
            elif result.status == IntegrationStatus.Unbounded:
                status = SolverStatus.Unbounded
                break
            elif result.status == IntegrationStatus.Penalty:
                # Continuation criterion is violated => update penalty parameter
                logger.debug(
                    "Updating penalty parameter %f -> %f", self.rho, 10 * self.rho
                )
                self.rho *= 10
                curr_filter = self.create_filter(curr_z, self.rho)

            if (params.iteration_limit is not None) and (
                iteration >= params.iteration_limit
            ):
                status = SolverStatus.IterationLimit
                logger.debug("Iteration limit reached")
                break

        (curr_x, curr_y) = self.flow.split_states(curr_z)
        iterate = Iterate(problem, params, curr_x, curr_y)

        x = iterate.x
        y = iterate.y
        d = iterate.bounds_dual

        iterations = iteration

        logger.debug("Status: %s", status)

        direct_dist = iterate.dist(initial_iterate)

        total_time = timer.elapsed()

        dist_factor = path_dist / direct_dist if direct_dist != 0.0 else 1.0

        self.print_result(total_time, status, iterate, iterations, dist_factor)

        (x, y, d) = self.transform.restore_sol(x, y, d)

        solver_result = SolverResult(
            problem,
            x,
            y,
            d,
            status,
            iterations=iterations,
            num_accepted_steps=accepted_steps,
            total_time=total_time,
            dist_factor=dist_factor,
        )

        if params.collect_path:
            solver_result._set_path(np.hstack(self.path), np.hstack(self.path_times))

        return solver_result



import abc

import numpy as np
import scipy as sp

from pygradflow.iterate import Iterate
from pygradflow.log import logger as lgg
from pygradflow.params import NewtonType, Params
from pygradflow.problem import Problem
from pygradflow.step.solver import step_solver
from pygradflow.step.solver.step_solver import StepResult, StepSolver

logger = lgg.getChild("newton")


class NewtonMethod(abc.ABC):
    def __init__(
        self, problem: Problem, orig_iterate: Iterate, dt: float, rho: float, tau=None
    ) -> None:
        self.problem = problem
        self.orig_iterate = orig_iterate
        self.dt = dt
        self.rho = rho
        self.tau = tau

    @property
    def params(self) -> Params:
        return self.orig_iterate.params

    @abc.abstractmethod
    def step(self, iterate: Iterate) -> StepResult:
        raise NotImplementedError()


class SimplifiedNewtonMethod(NewtonMethod):
    """
    Computes step based on the matrix given in terms of the *initial*
    iterate. Only requires a back-solve for each step.
    """

    def __init__(
        self,
        problem: Problem,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
        step_solver: StepSolver,
        tau=None,
    ) -> None:
        super().__init__(problem, orig_iterate, dt, rho, tau)
        self.func = step_solver.func

        self.step_solver = step_solver
        active_set = self.func.compute_active_set(orig_iterate, rho, tau)

        self.step_solver.update_active_set(active_set)
        self.step_solver.update_derivs(orig_iterate)

    def step(self, iterate):
        return self.step_solver.solve(iterate)


class FullNewtonMethod(NewtonMethod):
    """
    Computes step based on the matrix given in terms of the *current*
    iterate. Requires evaluation of the derivative and a factorization
    at each step.
    """

    def __init__(
        self,
        problem: Problem,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
        step_solver: StepSolver,
        tau=None,
    ) -> None:
        super().__init__(problem, orig_iterate, dt, rho, tau)
        self.func = step_solver.func
        self.step_solver = step_solver

    def step(self, iterate: Iterate) -> StepResult:
        active_set = self.func.compute_active_set(iterate, self.rho, self.tau)

        self.step_solver.update_active_set(active_set)
        self.step_solver.update_derivs(iterate)

        return self.step_solver.solve(iterate)


class FixedActiveSetNewtonMethod(NewtonMethod):
    """
    Computes step based on the matrix given in terms of the *current*
    iterate. Requires evaluation of the derivative and a factorization
    at each step.
    """

    def __init__(self, problem, active_set, orig_iterate, dt, rho, tau=None):
        super().__init__(problem, orig_iterate, dt, rho, tau)
        self.func = step_solver.func

        assert active_set.dtype == bool
        assert active_set.shape == problem.var_lb.shape

        self.active_set = active_set

        logger.info(
            "Active set fingerprint: %s, size: %s",
            hex(abs(hash(active_set.data.tobytes()))),
            np.sum(active_set),
        )

    def split_sol(self, s: np.ndarray):
        n = self.problem.num_vars
        m = self.problem.num_cons

        assert s.shape == (n + m,)
        return s[:n], s[n:]

    def create_step(self, iterate: Iterate, s: np.ndarray) -> StepResult:
        xn, yn = self.split_sol(s)
        x = iterate.x
        y = iterate.y

        dx = x - xn
        dy = y - yn

        return StepResult(iterate, dx, dy, active_set=self.active_set)

    @staticmethod
    def active_set_from_iterate(problem, iterate):
        lb = problem.var_lb
        ub = problem.var_ub

        x = iterate.x
        d = iterate.aug_lag_deriv_x()
        d_zero = np.isclose(d, 0.0)

        active_set = np.full(lb.shape, fill_value=True, dtype=bool)

        at_lower = np.isclose(x, lb)
        viol_lower = np.logical_and(x <= lb, np.logical_not(at_lower))
        active_set[viol_lower] = False

        d_neg = np.logical_and(d <= 0.0, np.logical_not(d_zero))
        active_set[np.logical_and(d_neg, at_lower)] = False

        at_upper = np.isclose(x, ub)
        viol_upper = np.logical_and(x >= ub, np.logical_not(at_upper))
        active_set[viol_upper] = False

        d_pos = np.logical_and(d >= 0.0, np.logical_not(d_zero))
        active_set[np.logical_and(d_pos, at_upper)] = False

        return active_set

    def step(self, iterate):
        rhs = self.func.value_at(iterate, self.rho, active_set=self.active_set)
        mat = self.func.deriv_at(iterate, self.rho, active_set=self.active_set)

        # logger.info(
        #     "Condition of system matrix: {0:.1e}".format(np.linalg.cond(mat.toarray()))
        # )

        solver = sp.sparse.linalg.splu(mat)

        s = solver.solve(rhs)
        next_step = self.create_step(iterate, s)
        next_iterate = next_step.iterate

        logger.info(
            "Initial rhs norm: {0}, final: {1}".format(
                np.linalg.norm(rhs), np.linalg.norm(self.func.value_at(next_iterate))
            )
        )

        return next_step


class ActiveSetNewtonMethod(NewtonMethod):
    """
    Computes step based on the matrix given in terms of the
    *initial* iterate with an active set based on the active set
    projection of the primal point of the *current* iterate. Requires a
    factorization at each step.
    """

    def __init__(
        self,
        problem: Problem,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
        step_solver: StepSolver,
        tau=None,
    ) -> None:
        super().__init__(problem, orig_iterate, dt, rho, tau)
        self.func = step_solver.func

        self.step_solver = step_solver
        self.step_solver.update_derivs(orig_iterate)

    def step(self, iterate):
        active_set = self.func.compute_active_set(iterate, self.rho, self.tau)

        self.step_solver.update_active_set(active_set)

        return self.step_solver.solve(iterate)


class GlobalizedNewtonMethod(NewtonMethod):
    """
    Globalized Newton method with Armijo line search. Globalization
    is based on the underlying function of the step solver.
    """

    def __init__(
        self,
        problem: Problem,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
        step_solver: StepSolver,
        tau=None,
    ) -> None:
        super().__init__(problem, orig_iterate, dt, rho, tau)
        self.func = step_solver.func
        self.step_solver = step_solver

    def _set_iterate(self, iterate):
        self.step_solver.update_derivs(iterate)
        active_set = self.func.compute_active_set(iterate, self.rho, self.tau)
        self.step_solver.update_active_set(active_set)

    def step(self, iterate):
        problem = self.problem
        params = iterate.params

        self._set_iterate(iterate)

        step_result = self.step_solver.solve(self.orig_iterate)

        # Armijo line search:
        alpha = 1.0

        func_value = self.func.value_at(iterate, self.rho)
        res_value = 0.5 * np.dot(func_value, func_value)

        # TODO: Create a method in the step function
        # to compute a forward product instead to make
        # everything more efficient
        func_deriv = self.func.deriv_at(iterate, self.rho)
        func_grad = func_deriv.T @ func_value

        func_grad_x = func_grad[: problem.num_vars]
        func_grad_y = func_grad[problem.num_vars :]

        dx = step_result.dx
        dy = step_result.dy

        inner_product = np.dot(func_grad_x, dx) + np.dot(func_grad_y, dy)

        max_it = 30

        for it in range(max_it):
            next_iterate = Iterate(
                problem, params, iterate.x - dx, iterate.y - dy, iterate.eval
            )

            next_func_value = self.func.value_at(next_iterate, self.rho)
            next_res_value = 0.5 * np.dot(next_func_value, next_func_value)

            if np.isclose(next_res_value, 0.0):
                break

            if next_res_value <= res_value + (1e-4 * alpha * inner_product):
                break

            alpha *= 0.5
            dx = alpha * step_result.dx
            dy = alpha * step_result.dy

        else:
            raise Exception("Line search failed to converge")

        logger.debug("Line search converged in %d iterations", it + 1)

        step_result = StepResult(self.orig_iterate, dx, dy, active_set=None, rcond=None)

        step_result.active_set = self.func.compute_active_set(
            step_result.iterate, self.rho, self.tau
        )

        return step_result


def newton_method(
    problem: Problem, params: Params, iterate: Iterate, dt: float, rho: float, tau=None
) -> NewtonMethod:
    assert dt > 0.0
    assert rho > 0.0

    solver = step_solver(problem, params, iterate, dt, rho)

    if params.newton_type == NewtonType.Simplified:
        return SimplifiedNewtonMethod(problem, iterate, dt, rho, solver, tau)
    elif params.newton_type == NewtonType.Full:
        return FullNewtonMethod(problem, iterate, dt, rho, solver, tau)
    elif params.newton_type == NewtonType.ActiveSet:
        return ActiveSetNewtonMethod(problem, iterate, dt, rho, solver, tau)
    else:
        assert params.newton_type == NewtonType.Globalized
        return GlobalizedNewtonMethod(problem, iterate, dt, rho, solver, tau)



import numpy as np
import scipy as sp

from .linear_solver import LinearSolver, LinearSolverError


class GMRESSolver(LinearSolver):
    def __init__(self, mat: sp.sparse.spmatrix, symmetric=False) -> None:
        super().__init__(mat, symmetric=symmetric)
        self.mat = mat

    def solve(self, rhs, trans=False, initial_sol=None):
        mat = self.mat.T if trans else self.mat

        if initial_sol is not None:
            initial_sol = initial_sol()

        (n, _) = mat.shape

        atol = 1e-8

        # Workaround for scipy bug
        if initial_sol is not None:
            res = rhs - mat @ initial_sol
            if np.linalg.norm(res, ord=np.inf) < atol:
                return initial_sol

        result = sp.sparse.linalg.gmres(mat, rhs, maxiter=n, x0=initial_sol, atol=atol)

        (sol, info) = result

        if info != 0:
            raise LinearSolverError("GMRES failed with error code {}".format(info))

        return sol



import math

import numpy as np
import scipy as sp

from pygradflow.linear_solver import LinearSolver
from pygradflow.log import logger
from pygradflow.params import Params

seed = 42


class ConditionEstimator:
    """
    Estimates condition of matrix using generic lienar solver.
    Based on
    "Estimating Extremal Eigenvalues and Condition Numbers of Matrices"
    by Dixon
    """

    def __init__(
        self,
        mat: sp.sparse.spmatrix,
        linear_solver: LinearSolver,
        params: Params,
        min_prob: float = 0.99,
        factor: float = 10.0,
    ):
        assert 0 < min_prob < 1
        self.size = mat.shape[0]

        self.params = params
        assert mat.shape == (self.size, self.size)

        self.min_prob = min_prob
        self.factor = factor
        self.mat = mat
        self.linear_solver = linear_solver
        self.rng = np.random.default_rng(seed=seed)

    def _required_its(self):
        factor = (1.0 - self.min_prob) / 1.6 * math.pow(self.size, -0.5)
        return -2 * math.ceil(math.log(factor, self.factor))

    def _random_vec(self):
        rng = self.rng
        size = self.size
        vec = rng.normal(size=size)

        while True:
            if (vec != 0.0).any():
                break
            vec = rng.normal(size=size)

        norm = np.linalg.norm(vec)
        return (vec / norm).astype(self.params.dtype)

    def estimate_rcond(self) -> float:
        mat = self.mat
        trans_mat = mat.T
        linear_solver = self.linear_solver

        num_its = self._required_its()
        assert num_its > 0

        x = self._random_vec()
        y = self._random_vec()

        xprod = np.copy(x)
        yprod = np.copy(y)

        # Factors to prevent over- or underflow
        xfac = 1.0
        yfac = 1.0

        logger.debug("Number of iterations: %s", num_its)

        for k in range(num_its):
            xprod = mat @ xprod
            xprod = trans_mat @ xprod

            yprod = linear_solver.solve(yprod, trans=True)
            yprod = linear_solver.solve(yprod)

            xnorm = float(np.linalg.norm(xprod))
            ynorm = float(np.linalg.norm(yprod))

            xfac *= xnorm
            xprod /= xnorm

            yfac *= ynorm
            yprod /= ynorm

            assert y.dot(yprod) > 0.0

        pow_fac = 1.0 / (2.0 * num_its)

        xdot = x.dot(xprod) * xfac
        xdot = math.pow(xdot, pow_fac)

        ydot = y.dot(yprod) * yfac
        ydot = math.pow(ydot, pow_fac)

        if np.isinf(xdot) or np.isinf(ydot):
            return 0.0

        cond = xdot * ydot

        if np.isinf(cond):
            return 0.0

        rcond = 1.0 / cond

        return rcond



import cyipopt
import numpy as np
import scipy as sp
from scipy.optimize import Bounds, minimize

from pygradflow.implicit_func import ImplicitFunc
from pygradflow.iterate import Iterate
from pygradflow.step.step_control import (
    StepController,
    StepControlResult,
    StepSolverError,
)


class BoxReducedProblem(cyipopt.Problem):
    """
    Solves the primal-dual proximally regularized box-reduced problem
     .. math::
        \\begin{align}
            \\min_{x \\in \\mathbb{R}^{n}, w \\in \\mathbb{R}^{n}} \\quad
            & f(x) + \\frac{\\rho}{2} \\|c(x)\\|^{2} + \\frac{\\lambda}{2} (\\| x - \\hat{x} \\|^{2} + \\| -\\frac{1}{\\lambda}x - \\hat{y} \\|^{2}) \\\\
            \\text{s.t.} \\quad & l^x \\leq x \\leq u^x
        \\end{align}

    whose optimum coincides with the implicit Euler step of the augmented Lagrangian
    of the original problem. The problem only consists of box constraints.
    """

    def __init__(self, step_controller, iterate, lamb, rho):
        self.iterate = iterate
        self.step_controller = step_controller
        self.lamb = lamb
        self.rho = rho
        self.problem = iterate.problem

        self.var_lb = self.problem.var_lb
        self.var_ub = self.problem.var_ub

        self.num_vars = self.problem.num_vars

        super(BoxReducedProblem, self).__init__(
            n=self.num_vars,
            m=0,
            lb=self.var_lb,
            ub=self.var_ub,
        )

    def objective(self, x):
        return self.step_controller.objective(self.iterate, x, self.lamb, self.rho)

    def gradient(self, x):
        return self.step_controller.gradient(self.iterate, x, self.lamb, self.rho)

    def constraint(self, x):
        return np.array([])

    def jacobian(self, x):
        return np.empty((0, self.num_vars))

    def hessianstructure(self):
        x0 = self.x0
        hess = self.step_controller.hessian(self.iterate, x0, self.lamb, self.rho)
        hess = hess.tocoo()
        rows = hess.row
        cols = hess.col

        hess_filter = rows >= cols

        return rows[hess_filter], cols[hess_filter]

    def hessian(self, x, lag, obj_factor):
        hess = self.step_controller.hessian(self.iterate, x, self.lamb, self.rho)
        hess = hess.tocoo()

        rows = hess.row
        cols = hess.col
        data = hess.data

        hess_filter = rows >= cols

        return data[hess_filter]

    def set_options(self, timer):
        import logging

        logging.getLogger("cyipopt").setLevel(logging.WARNING)
        self.add_option("print_level", 0)
        # self.add_option("derivative_test", "second-order")
        # self.add_option("hessian_approximation", "limited-memory")
        # self.add_option("max_iter", 10)

        remaining = timer.remaining()

        if remaining <= 0.0:
            raise StepSolverError("Time limit reached")
        elif np.isfinite(remaining):
            self.add_option("max_cpu_time", remaining)

    @property
    def x0(self):
        return self.iterate.x

    def solve(self, timer):
        self.set_options(timer)
        x0 = self.x0

        # Solve using Ipopt
        x, info = super().solve(x0)

        if info["status"] != 0:
            raise StepSolverError("Ipopt failed to solve the problem")

        return x


class BoxReducedController(StepController):

    def objective(self, iterate, x, lamb, rho):
        eval = iterate.eval
        xhat = iterate.x
        yhat = iterate.y

        obj = eval.obj(x)
        cons = eval.cons(x)

        dx = x - xhat
        dx_norm_sq = np.dot(dx, dx)

        w = -1 / lamb * cons
        dy = w - yhat
        dy_norm_sq = np.dot(dy, dy)

        curr_obj = obj + 0.5 * rho * np.dot(cons, cons)
        curr_obj += 0.5 * lamb * (dx_norm_sq + dy_norm_sq)

        return curr_obj

    def gradient(self, iterate, x, lamb, rho):
        eval = iterate.eval
        xhat = iterate.x
        yhat = iterate.y

        obj_grad = eval.obj_grad(x)
        cons = eval.cons(x)
        cons_jac = eval.cons_jac(x)

        dx = x - xhat
        cons_jac_factor = (rho + (1 / lamb)) * cons + yhat
        return obj_grad + lamb * dx + cons_jac.T.dot(cons_jac_factor)

    def hessian(self, iterate, x, lamb, rho):
        eval = iterate.eval
        (n,) = x.shape

        yhat = iterate.y

        cons = eval.cons(x)
        jac = eval.cons_jac(x)
        cons_factor = 1 / lamb + rho
        y = cons_factor * cons + yhat

        hess = eval.lag_hess(x, y)
        hess += sp.sparse.diags([lamb], shape=(n, n))
        hess += cons_factor * (jac.T @ jac)

        return hess

    def solve_step_scipy(self, iterate, rho, dt, timer):
        problem = self.problem

        x = iterate.x
        lamb = 1.0 / dt

        lb = problem.var_lb
        ub = problem.var_ub

        def opt_res(x):
            grad = gradient(x)

            at_lb = np.isclose(x, lb)
            at_ub = np.isclose(x, lb)

            grad[at_lb] = np.maximum(grad[at_lb], 0.0)
            grad[at_ub] = np.minimum(grad[at_ub], 0.0)

            return np.linalg.norm(grad, ord=np.inf)

        def objective(x):
            return self.objective(iterate, x, lamb, rho)

        def gradient(x):
            return self.gradient(iterate, x, lamb, rho)

        def callback(x):
            if timer.reached_time_limit():
                raise StopIteration()

        result = minimize(
            objective,
            x,
            jac=gradient,
            bounds=Bounds(lb, ub),
            method="TNC",
            options={"gtol": 1e-8, "ftol": 0.0},
            callback=callback,
        )

        if not result.success:
            raise StepSolverError("Scipy failed to solve the problem")

        return result.x

    def solve_step_ipopt(self, iterate, rho, dt, timer):
        lamb = 1.0 / dt
        reduced_problem = BoxReducedProblem(self, iterate, lamb, rho)
        return reduced_problem.solve(timer=timer)

    def solve_step_box(self, iterate, rho, dt, timer):
        from .box_solver import BoxSolverError, solve_box_constrained

        problem = self.problem
        lamb = 1.0 / dt

        def objective(x):
            return self.objective(iterate, x, lamb, rho)

        def gradient(x):
            return self.gradient(iterate, x, lamb, rho)

        def hessian(x):
            return self.hessian(iterate, x, lamb, rho)

        try:
            return solve_box_constrained(
                iterate.x, objective, gradient, hessian, problem.var_lb, problem.var_ub
            )
        except BoxSolverError as e:
            raise StepSolverError("Box-constrained solver failed to converge") from e

    def step(
        self, iterate, rho: float, dt: float, display: bool, timer
    ) -> StepControlResult:

        lamb = 1.0 / dt

        problem = self.problem
        params = self.params

        # x = self.solve_step_ipopt(iterate, rho, dt, timer)

        # Note: The minimize function shipped with scipy
        # do not consistently produce high-quality solutions,
        # causing the optimization of the overall problem to fail.
        # x = self.solve_step_scipy(iterate, rho, dt, timer)
        x = self.solve_step_box(iterate, rho, dt, timer)

        eval = iterate.eval
        cons = eval.cons(x)
        w = (-1 / lamb) * cons
        y = iterate.y - w

        next_iterate = Iterate(problem, params, x, y, iterate.eval)

        implicit_func = ImplicitFunc(problem, iterate, dt)

        value = implicit_func.value_at(next_iterate, rho)

        residuum = np.linalg.norm(value)

        if residuum <= 1e-6:
            return StepControlResult(
                next_iterate, 0.5 * lamb, active_set=None, rcond=None, accepted=True
            )

        return StepControlResult(
            next_iterate, 2.0 * lamb, active_set=None, rcond=None, accepted=False
        )
